SystemOrganization addCategory: #'ST80-Kernel-Remnants'!
SystemOrganization addCategory: #'ST80-Menus'!
SystemOrganization addCategory: #'ST80-Paths'!
SystemOrganization addCategory: #'ST80-Support'!

!String methodsFor: '*ST80' stamp: 'yo 11/3/2004 19:24'!
asParagraph
	"Answer a Paragraph whose text string is the receiver."

	^Paragraph withText: self asText! !

Object subclass: #ParagraphEditor
	instanceVariableNames: 'model paragraph startBlock stopBlock beginTypeInBlock emphasisHere initialText selectionShowing otherInterval lastParentLocation'
	classVariableNames: 'AutoIndent CaseSensitiveFinds ChangeText CmdActions CmdKeysInText FindText ShiftCmdActions UndoInterval UndoMessage UndoParagraph UndoSelection Undone WalkAlongDisplayedLine'
	poolDictionaries: 'TextConstants'
	category: 'ST80-Kernel-Remnants'!

!ParagraphEditor commentStamp: '<historical>' prior: 0!
I am a Controller for editing a Paragraph. I am a kind of ScrollController, so that more text can be created for the Paragraph than can be viewed on the screen. Editing messages are sent by issuing commands from a yellow button menu or from keys on the keyboard. My instances keep control as long as the cursor is within the view when the red or yellow mouse button is pressed; they give up control if the blue button is pressed.!

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'sw 5/27/2000 00:03'!
abandonChangeText
	"Call this to get out of the maddening situation in which the system keeps aggressively trying to do a replacement that you no longer wish to make, every time you make choose a new method in a list."
	ChangeText := FindText

	"ParagraphEditor abandonChangeText"
! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:50'!
autoIndent
	^ AutoIndent ifNil: [AutoIndent := true]! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:50'!
autoIndent: aBoolean
	AutoIndent := aBoolean! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 22:56'!
caseSensitiveFinds
	^ CaseSensitiveFinds ifNil: [CaseSensitiveFinds := false]! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 22:57'!
caseSensitiveFinds: aBoolean
	CaseSensitiveFinds := aBoolean! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:52'!
cmdKeysInText
	^ CmdKeysInText ifNil: [CmdKeysInText := true]! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:52'!
cmdKeysInText: aBoolean
	CmdKeysInText := aBoolean! !

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'michael.rueger 3/23/2009 17:51'!
initialize
	"Initialize the keyboard shortcut maps and the shared buffers 
	for copying text across views and managing again and undo. 
	Marked this method changed to trigger reinit"

	"ParagraphEditor initialize"

	UndoSelection := FindText := ChangeText := Text new.
	UndoMessage := Message selector: #halt.
	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.! !

!ParagraphEditor class methodsFor: 'keyboard shortcut tables' stamp: 'michael.rueger 2/23/2009 18:15'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"

	"ParagraphEditor initialize"

	| cmdMap |

	cmdMap := Array new: 256 withAll: #noop:.	"use temp in case of a crash"

	cmdMap at: 1 + 1 put: #cursorHome:.			"home key"
	cmdMap at: 4 + 1 put: #cursorEnd:.				"end key"
	cmdMap at: 8 + 1 put: #backspace:.				"ctrl-H or delete key"
	cmdMap at: 11 + 1 put: #cursorPageUp:.		"page up key"
	cmdMap at: 12 + 1 put: #cursorPageDown:.	"page down key"
	cmdMap at: 13 + 1 put: #crWithIndent:.			"cmd-Return"
	cmdMap at: 27 + 1 put: #offerMenuFromEsc:.	"escape key"
	cmdMap at: 28 + 1 put: #cursorLeft:.			"left arrow key"
	cmdMap at: 29 + 1 put: #cursorRight:.			"right arrow key"
	cmdMap at: 30 + 1 put: #cursorUp:.				"up arrow key"
	cmdMap at: 31 + 1 put: #cursorDown:.			"down arrow key"
	cmdMap at: 32 + 1 put: #selectWord:.			"space bar key"
	cmdMap at: 127 + 1 put: #forwardDelete:keyEvent:.		"del key"

	'0123456789-=' 
		do: [:char | cmdMap at: char asciiValue + 1 put: #changeEmphasis:keyEvent:].

	'([{''"<' do: [:char | cmdMap at: char asciiValue + 1 put: #enclose:keyEvent:].

	cmdMap at: $, asciiValue + 1 put: #shiftEnclose:keyEvent:.

	"triplet = {character. comment selector. novice appropiated}"
	#(
		($a		#selectAll:				true)
		($b		#browseIt:				false)
		($c		#copySelection:			true)
		($d		#doIt:						false)
		($e		#exchange:				true)
		($f		#find:						true)
		($g		#findAgain:				true)
		($h		#setSearchString:		true)
		($i		#inspectIt:				false)
		($j		#doAgainOnce:			true)
		($l		#cancel:					true)
		($m		#implementorsOfIt:		false)
		($n		#sendersOfIt:			false)
		($p		#printIt:					false)
		($q		#querySymbol:			false)
		($s		#save:					true)
		($v		#paste:					true)
		($w		#backWord:				true)
		($x		#cut:						true)
		($y		#swapChars:				true)
		($z		#undo:					true)
	)
		do: [:triplet | cmdMap at: triplet first asciiValue + 1 put: triplet second].

	CmdActions := cmdMap.
! !

!ParagraphEditor class methodsFor: 'keyboard shortcut tables' stamp: 'CB 7/27/2009 14:50'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	| cmdMap |

	"shift-command and control shortcuts"
	cmdMap := Array new: 256 withAll: #noop:.  "use temp in case of a crash"

	cmdMap at: ( 1 + 1) put: #cursorHome:.				"home key"
	cmdMap at: ( 4 + 1) put: #cursorEnd:.				"end key"
	cmdMap at: ( 8 + 1) put: #forwardDelete:keyEvent:.			"ctrl-H or delete key"
	cmdMap at: (11 + 1) put: #cursorPageUp:.			"page up key"
	cmdMap at: (12 + 1) put: #cursorPageDown:.		"page down key"
	cmdMap at: (13 + 1) put: #crWithIndent:.			"ctrl-Return"
	cmdMap at: (27 + 1) put: #offerMenuFromEsc:.	"escape key"
	cmdMap at: (28 + 1) put: #cursorLeft:.				"left arrow key"
	cmdMap at: (29 + 1) put: #cursorRight:.				"right arrow key"
	cmdMap at: (30 + 1) put: #cursorUp:.				"up arrow key"
	cmdMap at: (31 + 1) put: #cursorDown:.			"down arrow key"
	cmdMap at: (32 + 1) put: #selectWord:.				"space bar key"
	cmdMap at: (45 + 1) put: #changeEmphasis:keyEvent:.		"cmd-sh-minus"
	cmdMap at: (61 + 1) put: #changeEmphasis:keyEvent:.		"cmd-sh-plus"
	cmdMap at: (127 + 1) put: #forwardDelete:keyEvent:.		"del key"

	"Note: Command key overrides shift key, so, for example, cmd-shift-9 produces $9 not $("
	'9[,''' do: [ :char | cmdMap at: (char asciiValue + 1) put: #shiftEnclose:keyEvent: ].	"({< and double-quote"
	"Note: Must use cmd-9 or ctrl-9 to get '()' since cmd-shift-9 is a Mac FKey command."

	"NB: sw 12/9/2001 commented out the idiosyncratic line just below, which was grabbing shift-esc in the text editor and hence which argued with the wish to have shift-esc be a universal gesture for escaping the local context and calling up the desktop menu."  
	"cmdMap at: (27 + 1) put: #shiftEnclose:." 	"ctrl-["

	"'""''(' do: [ :char | cmdMap at: (char asciiValue + 1) put: #enclose:]."

	"triplet = {character. comment selector. novice appropiated}"
	#(
		($a		argAdvance:						false)
		($b		browseItHere:					false)
		($c		compareToClipboard:			false)
		($d		debugIt:						false)
		($e		methodStringsContainingIt:	false)
		($f		displayIfFalse:					false)
		($g		fileItIn:							false)
		($h		cursorTopHome:					true)
		($i		exploreIt:							false)
		($j		doAgainMany:					true)
		($k		changeStyle:						true)
		($l		outdent:							true)
		($m	selectCurrentTypeIn:			true)
		($n		referencesToIt:					false)
		($r		indent:							true)
		($s		search:							true)
		($t		displayIfTrue:					false)
		($u		changeLfToCr:					false)
		($v		pasteInitials:						false)
		($w	methodNamesContainingIt:	false)
		($x		makeLowercase:					true)
		($y		makeUppercase:					true)
		($z		makeCapitalized:				true)
	)
		do: [:triplet |
			cmdMap at: (triplet first asciiValue         + 1) put: triplet second.		"plain keys"
			cmdMap at: (triplet first asciiValue - 32 + 1) put: triplet second.		"shifted keys"
			cmdMap at: (triplet first asciiValue - 96 + 1) put: triplet second.		"ctrl keys"
		].

	ShiftCmdActions := cmdMap! !

!ParagraphEditor class methodsFor: 'keyboard shortcut tables' stamp: 'sps 7/24/2003 17:25'!
multiRedoOverride
"Call this to set meta-r to perform the multilevel redo (or tweak the code below to have it bound to some other key sequence)."

"
ParagraphEditor multiRedoOverride.
"
	CmdActions at: $r asciiValue + 1 put: #multiRedo: 
! !

!ParagraphEditor class methodsFor: 'instance creation' stamp: 'nk 9/3/2004 14:10'!
new
	"Answer a new instance of me with a null Paragraph to be edited."

	| aParagraphEditor |
	aParagraphEditor := super new.
	aParagraphEditor changeParagraph: '' asParagraph.
	^aParagraphEditor! !

!ParagraphEditor class methodsFor: 'instance creation' stamp: ''!
newParagraph: aParagraph 
	"Answer an instance of me with aParagraph as the text to be edited."

	| aParagraphEditor |
	aParagraphEditor := super new.
	aParagraphEditor initialize.
	aParagraphEditor changeParagraph: aParagraph.
	^aParagraphEditor! !

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'tbn 7/6/2010 16:15'!
shiftedYellowButtonMenu
	"Answer the menu to be presented when the yellow button is pressed while the shift key is down"

	^ MenuMorph fromArray: {
		
		{'Browse it (b)' translated.					#browseIt}.
		{'Senders of it (n)' translated.				#sendersOfIt}.
		{'Implementors of it (m)' translated.		#implementorsOfIt}.
		{'References to it (N)' translated.			#referencesToIt}.
		#-.
		{'Selectors containing it (W)' translated.	#methodNamesContainingIt}.
		{'Method strings with it (E)' translated.	#methodStringsContainingit}.
		{'Method source with it' translated.		#methodSourceContainingIt}.
		{'Class names containing it' translated.	#classNamesContainingIt}.
		{'Class comments with it' translated.		#classCommentsContainingIt}.
		{'Change sets with it' translated.			#browseChangeSetsWithSelector}.
	"	#-.
		{'Pretty print' translated.					#prettyPrint}.
		{'Pretty print with color' translated.		#prettyPrintWithColor}.
		{'File it in (G)' translated.					#fileItIn}.
		#-.
		{'Back...' translated.						#yellowButtonActivity}.
	"
	}
! !

!ParagraphEditor class methodsFor: 'keyboard shortcut tables' stamp: 'sbw 10/8/1999 21:42'!
specialShiftCmdKeys

"Private - return array of key codes that represent single keys acting
as if shift-command were also being pressed"

^#(
	1	"home"
	3	"enter"
	4	"end"
	8	"backspace"
	11	"page up"
	12	"page down"
	27	"escape"
	28	"left arrow"
	29	"right arrow"
	30	"up arrow"
	31	"down arrow"
	127	"delete"
	)! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 13:37'!
walkAlongDisplayedLine
	^ WalkAlongDisplayedLine ifNil: [WalkAlongDisplayedLine := true]! !

!ParagraphEditor class methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 13:37'!
walkAlongDisplayedLine: aBoolean
	WalkAlongDisplayedLine := aBoolean! !

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'tbn 7/6/2010 16:16'!
yellowButtonExpertMenu
	^ MenuMorph fromArray: {
			{'Do it (d)' translated.					#doIt}.
			{'Print it (p)' translated.					#printIt}.
			{'Inspect it (i)' translated.				#inspectIt}.
			{'Explore it (I)' translated.				#exploreIt}.
			{'Debug it (D)' translated.				#debugIt}.
			{'Profile it' translated.						#tallyIt}.
		     {'Watch it' translated.					#watchIt}.	 
		    	 
			#-.
			{'Find...(f)' translated.					#find}.
			{'Find again (g)' translated.				#findAgain}.
			{'Extended search...' translated.			#shiftedTextPaneMenuRequest}.
			#-.
			{'Do again (j)' translated.				#again}.
			{'Undo (z)' translated.					#undo}.
			#-.
			{'Copy (c)' translated.					#copySelection}.
			{'Cut (x)' translated.						#cut}.
			{'Paste (v)' translated.					#paste}.
			{'Paste...' translated.					#pasteRecent}.
			#-.
			{'Accept (s)' translated.					#accept}.
			{'Cancel (l)' translated.					#cancel}.
		}.
! !

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'marcus.denker 11/19/2008 13:49'!
yellowButtonMenu

	^self yellowButtonExpertMenu
! !

!ParagraphEditor class methodsFor: 'class initialization' stamp: 'stephane.ducasse 9/13/2008 16:40'!
yellowButtonNoviceMenu

	^ MenuMorph fromArray: {
			{'set font... (k)' translated.				#offerFontMenu}.
			{'set style... (K)' translated.				#changeStyle}.
			{'set alignment... (u)' translated.		#chooseAlignment}.
			#-.
			{'find...(f)' translated.					#find}.
			{'find again (g)' translated.				#findAgain}.
			{'set search string (h)' translated.		#setSearchString}.
			#-.
			{'do again (j)' translated.				#again}.
			{'undo (z)' translated.					#undo}.
			#-.
			{'copy (c)' translated.					#copySelection}.
			{'cut (x)' translated.						#cut}.
			{'paste (v)' translated.					#paste}.
			{'paste...' translated.					#pasteRecent}.
			#-.
			{'accept (s)' translated.					#accept}.
			{'cancel (l)' translated.					#cancel}.
		}.
! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'jm 5/3/1998 19:19'!
accept
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling."

	initialText := paragraph text copy.
! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/18/2002 14:37'!
adjustSelection: directionBlock
	"Helper function for Cursor movement. Always moves point thus allowing selections to shrink. "
	"See also expandSelection:"
	"Accepts a one argument Block that computes the new postion given an old one."
	| newPosition |
	newPosition := directionBlock value: self pointIndex.
	self selectMark: self markIndex point: newPosition.
	^true.! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 10/28/2003 12:11'!
afterSelectionInsertAndSelect: aString

	self insertAndSelect: aString at: self stopIndex ! !

!ParagraphEditor methodsFor: 'menu messages' stamp: ''!
again
	"Text substitution. If the left shift key is down, the substitution is made 
	throughout the entire Paragraph. Otherwise, only the next possible 
	substitution is made.
	Undoer & Redoer: #undoAgain:andReselect:typedKey:."

	"If last command was also 'again', use same keys as before"
	self againOrSame: (UndoMessage sends: #undoAgain:andReselect:typedKey:)! !

!ParagraphEditor methodsFor: 'private' stamp: 'AlainPlantec 12/10/2009 23:04'!
againOnce: indices
	"Find the next occurrence of FindText.  If none, answer false.
	Append the start index of the occurrence to the stream indices, and, if
	ChangeText is not the same object as FindText, replace the occurrence by it.
	Note that the search is case-sensitive for replacements, otherwise not."

	| where |
	where := paragraph text findString: FindText startingAt: self stopIndex
				caseSensitive: ((ChangeText ~~ FindText) or: [self caseSensitiveFinds]).
	where = 0 ifTrue: [^ false].
	self deselect; selectInvisiblyFrom: where to: where + FindText size - 1.
	ChangeText ~~ FindText ifTrue: [self zapSelectionWith: ChangeText].
	indices nextPut: where.
	self selectAndScroll.
	^ true! !

!ParagraphEditor methodsFor: 'private' stamp: 'michael.rueger 2/23/2009 13:23'!
againOrSame: useOldKeys
	"Subroutine of search: and again.  If useOldKeys, use same FindText and ChangeText as before.
	 1/26/96 sw: real worked moved to againOrSame:many:"

	^ self againOrSame: useOldKeys many: Sensor leftShiftDown! !

!ParagraphEditor methodsFor: 'private' stamp: 'PeterHugossonMiller 9/3/2009 10:13'!
againOrSame: useOldKeys many: many
	"Subroutine of search: and again.  If useOldKeys, use same FindText and ChangeText as before.  If many is true, do it repeatedly.  Created 1/26/96 sw by adding the many argument to #againOrSame."

	|  home indices wasTypedKey |

	home := self selectionInterval.  "what was selected when 'again' was invoked"

	"If new keys are to be picked..."
	useOldKeys ifFalse: "Choose as FindText..."
		[FindText := UndoSelection.  "... the last thing replaced."
		"If the last command was in another paragraph, ChangeText is set..."
		paragraph == UndoParagraph ifTrue: "... else set it now as follows."
			[UndoInterval ~= home ifTrue: [self selectInterval: UndoInterval]. "blink"
			ChangeText := ((UndoMessage sends: #undoCutCopy:) and: [self hasSelection])
				ifTrue: [FindText] "== objects signal no model-locking by 'undo copy'"
				ifFalse: [self selection]]]. "otherwise, change text is last-replaced text"

	(wasTypedKey := FindText size = 0)
		ifTrue: "just inserted at a caret"
			[home := self selectionInterval.
			self replaceSelectionWith: self nullText.  "delete search key..."
			FindText := ChangeText] "... and search for it, without replacing"
		ifFalse: "Show where the search will start"
			[home last = self selectionInterval last ifFalse:
				[self selectInterval: home]].

	"Find and Change, recording start indices in the array"
	indices := (Array new: 20) writeStream. "an array to store change locs"
	[(self againOnce: indices) & many] whileTrue. "<-- this does the work"
	indices isEmpty ifTrue:  "none found"
		[self flash.
		wasTypedKey ifFalse: [^self]].

	(many | wasTypedKey) ifFalse: "after undo, select this replacement"
		[home := self startIndex to:
			self startIndex + UndoSelection size - 1].

	self undoer: #undoAgain:andReselect:typedKey: with: indices contents with: home with: wasTypedKey! !

!ParagraphEditor methodsFor: 'menu messages' stamp: ''!
align
	"Align text according to the next greater alignment value--cycling among 
	left flush, right flush, center, justified.  No effect on the undoability of the pre
	preceding command."

	paragraph toggleAlignment.
	paragraph displayOn: Display.
	self recomputeInterval! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:20'!
align: characterStream 
	"Triggered by Cmd-u;  cycle through alignment alternatives.  8/11/96 sw"

	self align.
	^ true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:21'!
argAdvance: characterStream
	"Invoked by Ctrl-a.  Useful after Ctrl-q.
	 Search forward from the end of the selection for a colon followed by
		a space.  Place the caret after the space.  If none are found, place the
		caret at the end of the text.  Does not affect the undoability of the 
	 	previous command."

	| start |
	self closeTypeIn: characterStream.
	start := paragraph text findString: ': ' startingAt: self stopIndex.
	start = 0 ifTrue: [start := paragraph text size + 1].
	self selectAt: start + 2.
	^true! !

!ParagraphEditor methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:50'!
autoIndent
	^ self class autoIndent! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'yo 3/16/2004 13:05'!
backTo: startIndex
	"During typing, backspace to startIndex.  Deleted characters fall into three
	 clusters, from left to right in the text: (1) preexisting characters that were
	 backed over; (2) newly typed characters that were backed over (excluding
	 typeahead, which never even appears); (3) preexisting characters that
	 were highlighted before typing began.  If typing has not yet been opened,
	 open it and watch for the first and third cluster.  If typing has been opened,
	 watch for the first and second cluster.  Save characters from the first and third
	 cluster in UndoSelection.  Tally characters from the first cluster in UndoMessage's parameter.
	 Delete all the clusters.  Do not alter Undoer or UndoInterval (except via
	 openTypeIn).  The code is shorter than the comment."

	| saveLimit newBackovers |
	saveLimit := beginTypeInBlock == nil
		ifTrue: [self openTypeIn. UndoSelection := self nullText. self stopIndex]
		ifFalse: [self startOfTyping].
	self setMark: startIndex.
	startIndex < saveLimit ifTrue:
		[newBackovers := self startOfTyping - startIndex.
		beginTypeInBlock := self startIndex.
		UndoSelection replaceFrom: 1 to: 0 with:
			(paragraph text copyFrom: startIndex to: saveLimit - 1).
		UndoMessage argument: (UndoMessage argument ifNil: [1]) + newBackovers].
	self zapSelectionWith: self nullText.
	self unselect! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:21'!
backWord: characterStream 
	"If the selection is not a caret, delete it and leave it in the backspace buffer.
	 Else if there is typeahead, delete it.
	 Else, delete the word before the caret."

	| startIndex |
	characterStream isEmpty
		ifTrue:
			[self hasCaret
				ifTrue: "a caret, delete at least one character"
					[startIndex := 1 max: self markIndex - 1.
					[startIndex > 1 and:
						[(paragraph text at: startIndex - 1) asCharacter tokenish]]
						whileTrue:
							[startIndex := startIndex - 1]]
				ifFalse: "a non-caret, just delete it"
					[startIndex := self markIndex].
			self backTo: startIndex]
		ifFalse:
			[characterStream reset].
	^false! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 18:15'!
backspace: characterStream 
	"Backspace over the last character."

	| startIndex |
	Sensor leftShiftDown ifTrue: [^ self backWord: characterStream].
	characterStream isEmpty
		ifTrue:
			[startIndex := self markIndex + (self hasCaret ifTrue: [0] ifFalse: [1]).
			startIndex := 1 max: startIndex - 1.
			self backTo: startIndex]
		ifFalse:
			[characterStream skip: -1].
	^false! !

!ParagraphEditor methodsFor: 'parenblinking' stamp: 'mir 8/3/2004 13:31'!
blinkParenAt: parenLocation 
	self text
		addAttribute: TextEmphasis bold
		from: parenLocation
		to: parenLocation.
	lastParentLocation := parenLocation.! !

!ParagraphEditor methodsFor: 'parenblinking' stamp: 'michael.rueger 2/23/2009 13:31'!
blinkPrevParen: openDelimiter
	| closeDelimiter level string here hereChar |
	string := paragraph text string.
	here := startBlock stringIndex.
	closeDelimiter := '([{' at: (')]}' indexOf: openDelimiter).
	level := 1.
	[level > 0 and: [here > 2]]
		whileTrue:
			[hereChar := string at: (here := here - 1).
			hereChar = closeDelimiter
				ifTrue:
					[level := level - 1.
					level = 0
						ifTrue: [^ self blinkParenAt: here]]
				ifFalse:
					[hereChar = openDelimiter
						ifTrue: [level := level + 1]]].! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 12:57'!
browseIt
	"Launch a browser for the class indicated by the current selection. 
	If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	self lineSelectAndEmptyCheck: [^ self].
	aClass := SystemNavigation default
				classFromPattern: (self selection string copyWithoutAll: CharacterSet crlf)
				withCaption: 'choose a class to browse...'.
	aClass ifNil: [^ self internalCallToImplementorsOfIt].
	self terminateAndInitializeAround: [ | aBrow |
			aBrow := SystemBrowser default new.
			aBrow setClass: aClass selector: nil.
			aBrow class
				openBrowserView: (aBrow openEditString: nil)
				label: 'System Browser']! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'sd 5/24/2010 22:32'!
browseIt: characterStream 
	"Triggered by Cmd-B; browse the thing represented by the current selection, if plausible."

	self browseIt.
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:31'!
browseItHere: characterStream 
	"Triggered by Cmd-shift-B; browse the thing represented by the current selection, if plausible, in the receiver's own window.  3/1/96 sw"

	self browseItHere.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 6/11/2008 15:00'!
cancel 
	"Restore the text of the paragraph to be the text saved since initialization 
	or the last accept.  Undoer & Redoer: undoAndReselect:redoAndReselect:.
	This used to call controlTerminate and controlInitialize but this seemed illogical.
	Sure enough, nobody overrode them who had cancel in the menu, and if
	anybody really cared they could override cancel."

	UndoSelection := paragraph text.
	self undoer: #undoAndReselect:redoAndReselect: with: self selectionInterval with: (1 to: 0).
	self changeParagraph: (paragraph text: initialText).
	UndoParagraph := paragraph.
	otherInterval := UndoInterval := 1 to: initialText size. "so undo will replace all"
	paragraph displayOn: Display.
	self selectAt: 1.
	self scrollToTop
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:31'!
cancel: characterStream 
	"Cancel unsubmitted changes.  Flushes typeahead.  1/12/96 sw
	 1/22/96 sw: put in control terminate/init"

	self terminateAndInitializeAround: [self cancel].
	^ true! !

!ParagraphEditor methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 23:04'!
caseSensitiveFinds
	^ self class caseSensitiveFinds! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 2/6/2009 17:12'!
changeAlignment
	| aList reply  |
	aList := #(leftFlush centered justified rightFlush).
	reply := UIManager default chooseFrom: (aList collect: [:t | t translated]) values: aList.
	reply ifNil:[^self].
	self setAlignment: reply.
	paragraph composeAll.
	self recomputeSelection.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 2/6/2009 17:14'!
changeEmphasis
	| aList reply  |
	aList := #(normal bold italic narrow underlined struckOut).
	reply := UIManager default chooseFrom: (aList collect: [:t | t translated]) values: aList.
	reply ifNotNil: [
		self setEmphasis: reply.
		paragraph composeAll.
		self recomputeSelection].
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 13:48'!
changeEmphasis: characterStream keyEvent: keyEvent
	"Change the emphasis of the current selection or prepare to
	accept characters with the change in emphasis. Emphasis
	change amounts to a font change. Keeps typeahead."
	
	| keyCode attribute oldAttributes index thisSel colors extras |

	"control 0..9 -> 0..9"
	keyCode := ('0123456789-=' indexOf: keyEvent keyCharacter ifAbsent: [1]) - 1.

	oldAttributes := paragraph text attributesAt: self pointIndex forStyle: paragraph textStyle.
	thisSel := self selection.

	"Decipher keyCodes for Command 0-9..."
	(keyCode between: 1 and: 5) ifTrue: [
		attribute := TextFontChange fontNumber: keyCode
	].

	keyCode = 6 ifTrue: [
		| labels lines | 
		colors := #(#black #magenta #red #yellow #green #blue #cyan #white ).
		extras := (self class name = #TextMorphEditor and: [(self morph isKindOf: TextMorphForEditView) not])
						ifTrue: ["not a system window" #()]
						ifFalse: [#('Link to comment of class' 'Link to definition of class' 'Link to hierarchy of class' 'Link to method' )].

		
		labels := colors , #('choose color...' 'Do it' 'Print it' ) , extras , #('be a web URL link' 'Edit hidden info' 'Copy hidden info' ).
		lines := Array with: colors size + 1.
		index := UIManager default chooseFrom: labels lines: lines.
		index = 0 ifTrue: [ ^ true].
			
		index <= colors size ifTrue: [
			attribute := TextColor color: (Color perform: (colors at: index))
		]
		ifFalse: [
			index := index - colors size - 1. "Re-number!!!!!!"

			index = 0 ifTrue: [
				attribute := self chooseColor
			].

			index = 1 ifTrue: [
				attribute := TextDoIt new.
				thisSel := attribute analyze: self selection asString
			].

			index = 2 ifTrue: [
				attribute := TextPrintIt new.
				thisSel := attribute analyze: self selection asString
			].

			extras size = 0 & (index > 2) ifTrue: [
				index := index + 4 "skip those"
			].

			index = 3 ifTrue: [
				attribute := TextLink new.
				thisSel := attribute analyze: self selection asString with: 'Comment'
			].

			index = 4 ifTrue: [
				attribute := TextLink new.
				thisSel := attribute analyze: self selection asString with: 'Definition'
			].

			index = 5 ifTrue: [
				attribute := TextLink new.
				thisSel := attribute analyze: self selection asString with: 'Hierarchy'
			].

			index = 6 ifTrue: [
				attribute := TextLink new.
				thisSel := attribute analyze: self selection asString
			].
		
			index = 7 ifTrue: [
				attribute := TextURL new.
				thisSel := attribute analyze: self selection asString
			].
		
			index = 8 ifTrue: [
				"Edit hidden info"
				thisSel := self hiddenInfo. "includes selection"
				attribute := TextEmphasis normal
			].

			index = 9 ifTrue: [
				"Copy hidden info"
				self copyHiddenInfo.
				^ true
			].
		
			"no other action"
			thisSel
				ifNil: [ ^ true ]
		]
	].

	(keyCode between: 7 and: 11) ifTrue: [
		keyEvent leftShiftDown ifTrue: [
			keyCode = 10 ifTrue: [
				attribute := TextKern kern: -1
			].
			keyCode = 11 ifTrue: [
				attribute := TextKern kern: 1
			]
		]
		ifFalse: [
			attribute := TextEmphasis perform: (#(#bold #italic #narrow #underlined #struckOut ) at: keyCode - 6).
			oldAttributes
						do: [:att | (att dominates: attribute) ifTrue: [attribute turnOff]]
		]
	].

	keyCode = 0
		ifTrue: [attribute := TextEmphasis normal].

	beginTypeInBlock ~~ nil ifTrue: [
		"only change emphasisHere while typing"
		self insertTypeAhead: characterStream.
		emphasisHere := Text addAttribute: attribute toArray: oldAttributes.
		^ true
	].

	self
		replaceSelectionWith: (thisSel asText addAttribute: attribute).
		
	^ true
! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 2/6/2009 17:15'!
changeEmphasisOrAlignment
	| aList reply  |
	aList := #(normal bold italic narrow underlined struckOut leftFlush centered rightFlush justified).
	reply := UIManager default chooseFrom: (aList collect: [:t | t translated]) values: aList lines: #(6).
	reply ~~ nil ifTrue:
		[(#(leftFlush centered rightFlush justified) includes: reply)
			ifTrue:
				[paragraph perform: reply.
				self recomputeInterval]
			ifFalse:
				[self setEmphasis: reply.
				paragraph composeAll.
				self recomputeSelection.
				]].
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'nice 11/30/2009 22:31'!
changeLfToCr: characterStream 
	"Replace all LFs by CRs, and CR-LF pairs by single CRs.
	Triggered by Cmd-U -- useful when getting code from FTP sites"

	self replaceSelectionWith: (Text fromString:
		self selection string withSqueakLineEndings).
	^ true! !

!ParagraphEditor methodsFor: 'initialize-release' stamp: ''!
changeParagraph: aParagraph 
	"Install aParagraph as the one to be edited by the receiver."

	UndoParagraph == paragraph ifTrue: [UndoParagraph := nil].
	paragraph := aParagraph.
	self resetState! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 2/6/2009 17:16'!
changeStyle
	"Let user change styles for the current text pane  
	 Moved from experimentalCommand to its own method  "

	| aList reply style |
	aList := StrikeFont actualFamilyNames.
	aList addFirst: 'DefaultTextStyle'.
	reply := UIManager default chooseFrom: aList values: aList lines: #(1).
	reply ifNotNil:
		[(style := TextStyle named: reply) ifNil: [Beeper beep. ^ true].
		paragraph textStyle: style copy.
		paragraph composeAll.
		self recomputeSelection.
		].
	^ true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:32'!
changeStyle: characterStream 
	"Put up the style-change menu"

	self closeTypeIn: characterStream.
	self changeStyle.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'RAA 3/15/2001 12:10'!
changeStyleTo: aNewStyle

	paragraph textStyle: aNewStyle.
	paragraph composeAll.
	self recomputeSelection.
! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sw 9/27/1999 11:54'!
chooseAlignment
	self changeAlignment! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'tk 5/7/2001 09:11'!
chooseColor
	"Make a new Text Color Attribute, let the user pick a color, and return the attribute.  This is the non-Morphic version."

	^ TextColor color: (Color fromUser)! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 4/15/2003 22:40'!
classCommentsContainingIt
	"Open a browser class comments which contain the current selection somewhere in them."

	self lineSelectAndEmptyCheck: [^ self].
	self terminateAndInitializeAround: [
		self systemNavigation browseClassCommentsWithString: self selection string]! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'dvf 8/23/2003 11:51'!
classNamesContainingIt
	"Open a browser on classes whose names contain the selected string"

	self lineSelectAndEmptyCheck: [^self].
	self systemNavigation 
		browseClassesWithNamesContaining: self selection string
		caseSensitive: Sensor leftShiftDown! !

!ParagraphEditor methodsFor: 'parenblinking' stamp: 'mir 8/3/2004 13:31'!
clearParens
	lastParentLocation ifNotNil:
		[self text string size >= lastParentLocation ifTrue: [
			self text
				removeAttribute: TextEmphasis bold
				from: lastParentLocation
				to: lastParentLocation]]
! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'ar 1/15/2001 18:37'!
clipboardText

	^ Clipboard clipboardText! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'ar 1/15/2001 18:38'!
clipboardText: text

	^ Clipboard clipboardText: text! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'ar 1/15/2001 18:38'!
clipboardTextPut: text

	^ Clipboard clipboardText: text! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'th 9/19/2002 17:40'!
closeTypeIn
	"See comment in openTypeIn.  It is important to call closeTypeIn before executing
	 any non-typing key, making a new selection, etc.  It is called automatically for
	 menu commands.
	 Typing commands can call 'closeTypeIn: aCharacterStream' instead of this to
	 save typeahead.  Undoer & Redoer: undoAndReselect:redoAndReselect:."

	| begin stop |
	beginTypeInBlock == nil ifFalse:
		[(UndoMessage sends: #noUndoer) ifTrue: "should always be true, but just in case..."
			[begin := self startOfTyping.
			stop := self stopIndex.
			self undoer: #undoAndReselect:redoAndReselect:
				with: (begin + UndoMessage argument to: begin + UndoSelection size - 1)
				with: (stop to: stop - 1).
			UndoInterval := begin to: stop - 1].
		beginTypeInBlock := nil]! !

!ParagraphEditor methodsFor: 'typing support' stamp: ''!
closeTypeIn: characterStream
	"Call instead of closeTypeIn when you want typeahead to be inserted before the
	 control character is executed, e.g., from Ctrl-V."

	self insertTypeAhead: characterStream.
	self closeTypeIn! !

!ParagraphEditor methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:52'!
cmdKeysInText
	^ self class cmdKeysInText! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 1/21/2000 18:51'!
comment
	"All key actions that are neither editing nor typing actions have to
	send closeTypeIn at first. See comment in openTypeIn closeTypeIn"! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'di 11/23/1998 15:21'!
compareToClipboard
	"Check to see if whether the receiver's text is the same as the text currently on the clipboard, and inform the user."
	| s1 s2 |
	s1 := self clipboardText string.
	s2 := paragraph text string.
	s1 = s2 ifTrue: [^ self inform: 'Exact match'].

	(StringHolder new textContents:
		(TextDiffBuilder buildDisplayPatchFrom: s1 to: s2))
		openLabel: 'Comparison to Clipboard Text'! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:32'!
compareToClipboard: characterStream 
	"Compare the receiver to the text on the clipboard.  Flushes typeahead.  5/1/96 sw"

	self terminateAndInitializeAround: [self compareToClipboard].
	^ true! !

!ParagraphEditor methodsFor: 'watchIt' stamp: 'hfm 1/11/2009 11:38'!
compileSelectionAsBlock
	"Treat the current selection as an expression; evaluate it and return the result"
	
	| stream result rcvr ctxt |
	self lineSelectAndEmptyCheck: [^ ''].

	stream := ReadWriteStream on: (String new: (startBlock stringIndex) - (stopBlock stringIndex) + 2).
	stream nextPutAll: '[[ '.
	stream next: ((startBlock stringIndex) - (stopBlock stringIndex))
		putAll: paragraph string
		startingAt: stopBlock stringIndex.
	stream nextPutAll: ' ] on: Error do: [ :ex | ex ]]'.
	stream reset.

	(model respondsTo: #doItReceiver) 
		ifTrue: [ FakeClassPool adopt: model selectedClass.  "Include model pool vars if any"
				rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: stream
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [ FakeClassPool adopt: nil. 
					^ #failedDoit]
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	FakeClassPool adopt: nil.
	^ result! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'Igor.Stasenko 12/20/2009 07:26'!
compileSelectionFor: anObject in: evalContext

	| methodNode |
	methodNode := [Compiler new
		compileNoPattern: self selectionAsStream
		in: anObject class
		context: evalContext
		notifying: self
		ifFail: [^nil]]
			on: OutOfScopeNotification
			do: [:ex | ex resume: true].

	^ methodNode generateWithTempNames.
! !

!ParagraphEditor methodsFor: 'private' stamp: 'PeterHugossonMiller 9/3/2009 10:14'!
completeSymbol: hintText lastOffering: selectorOrNil
	"Invoked by Ctrl-q when there is only a caret.
		Do selector-completion, i.e., try to replace the preceding identifier by a
		selector that begins with those characters & has as many keywords as possible.
	 	Leave two spaces after each colon (only one after the last) as space for
		arguments.  Put the caret after the space after the first keyword.  If the
		user types Ctrl-q again immediately, choose a different selector.
	 Undoer: #undoQuery:lastOffering:; Redoer: itself.
	If redoing, just redisplay the last offering, selector[OrNil]."

	| firstTime input prior caret newStart sym kwds outStream |
	firstTime := self isRedoing
		ifTrue: [prior := sym := selectorOrNil. true]
		ifFalse: [hintText isNil].
	firstTime
		ifTrue: "Initial Ctrl-q (or redo)"					
			[caret := self startIndex.
			self selectPrecedingIdentifier.
			input := self selection]
		ifFalse: "Repeated Ctrl-q"
			[caret := UndoInterval first + hintText size.
			self selectInvisiblyFrom: UndoInterval first to: UndoInterval last.
			input := hintText.
			prior := selectorOrNil].
	(input size ~= 0 and: [sym ~~ nil or:
			[(sym := Symbol thatStarts: input string skipping: prior) ~~ nil]])
		ifTrue: "found something to offer"
			[newStart := self startIndex.
			outStream := (String new: 2 * sym size) writeStream.
			1 to: (kwds := sym keywords) size do:
				[:i |
				outStream nextPutAll: (kwds at: i).
				i = 1 ifTrue: [caret := newStart + outStream contents size + 1].
				outStream nextPutAll:
					(i < kwds size ifTrue: ['  '] ifFalse: [' '])].
			UndoSelection := input.
			self deselect; zapSelectionWith: outStream contents asText.
			self undoer: #undoQuery:lastOffering: with: input with: sym]
		ifFalse: "no more matches"
			[firstTime ifFalse: "restore original text & set up for a redo"
				[UndoSelection := self selection.
				self deselect; zapSelectionWith: input.
				self undoer: #completeSymbol:lastOffering: with: input with: prior.
				Undone := true].
			self flash].
	self selectAt: caret! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/17/2002 16:11'!
computeIntervalFrom: start to: stop
	"Select the designated characters, inclusive.  Make no visual changes."

	self setMark: start.
	self setPoint: stop + 1.! !

!ParagraphEditor methodsFor: 'controlling' stamp: ''!
controlInitialize

	super controlInitialize.
	self recomputeInterval.
	self initializeSelection.
	beginTypeInBlock := nil! !

!ParagraphEditor methodsFor: 'controlling' stamp: ''!
controlTerminate

	self closeTypeIn.  "Must call to establish UndoInterval"
	super controlTerminate.
	self deselect! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'tk 5/7/2001 08:47'!
copyHiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden
info.  Copy that to the clipboard.  You can paste it and see what it is.
Usually enclosed in <>."

	^ self clipboardTextPut: self hiddenInfo asText! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sw 8/1/97 15:09'!
copySelection
	"Copy the current selection and store it in the paste buffer, unless a caret.  Undoer & Redoer: undoCutCopy"

	self lineSelectAndEmptyCheck: [^ self].

	"Simulate 'substitute: self selection' without locking the controller"
	UndoSelection := self selection.
	self undoer: #undoCutCopy: with: self clipboardText.
	UndoInterval := self selectionInterval.
	self clipboardTextPut: UndoSelection! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:32'!
copySelection: characterStream 
	"Copy the current text selection.  Flushes typeahead."

	self copySelection.
	^true! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'di 5/6/1998 15:21'!
correctFrom: start to: stop with: aString
	"Make a correction in the model that the user has authorised from somewhere else in the system (such as from the compilier).  The user's selection is not changed, only corrected."
	| wasShowing userSelection delta loc |
	aString = '#insert period' ifTrue:
		[loc := start.
		[(loc := loc-1)>0 and: [(paragraph text string at: loc) isSeparator]]
			whileTrue: [loc := loc-1].
		^ self correctFrom: loc+1 to: loc with: '.'].
	(wasShowing := selectionShowing) ifTrue: [ self reverseSelection ].
	userSelection := self selectionInterval.

	self selectInvisiblyFrom: start to: stop.
	self replaceSelectionWith: aString asText.

	delta := aString size - (stop - start + 1).
	self selectInvisiblyFrom:
		userSelection first + (userSelection first > start ifFalse: [ 0 ] ifTrue: [ delta ])
		to: userSelection last + (userSelection last > start ifFalse: [ 0 ] ifTrue: [ delta ]).
	wasShowing ifTrue: [ self reverseSelection ].
! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:32'!
crWithIndent: characterStream 
	"Replace the current text selection with CR followed by as many tabs
	as on the current line (+/- bracket count) -- initiated by Shift-Return."
	| char s i tabCount |
	s := paragraph string.
	i := self stopIndex.
	tabCount := 0.
	[(i := i-1) > 0 and: [(char := s at: i) ~= Character cr]]
		whileTrue:  "Count tabs and brackets (but not a leading bracket)"
		[(char = Character tab and: [i < s size and: [(s at: i+1) ~= $[ ]]) ifTrue: [tabCount := tabCount + 1].
		char = $[ ifTrue: [tabCount := tabCount + 1].
		char = $] ifTrue: [tabCount := tabCount - 1]].
	characterStream crtab: tabCount.  "Now inject CR with tabCount tabs"
	^ false! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 11/18/2002 17:08'!
cursorDown: characterStream 

	"Private - Move cursor from position in current line to same position in
	next line. If next line too short, put at end. If shift key down,
	select."
	self closeTypeIn: characterStream.
	self 
		moveCursor:[:position | self
				sameColumn: position
				newLine:[:line | line + 1]
				forward: true]
		forward: true
		specialBlock:[:dummy | dummy].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 12/11/2009 13:38'!
cursorEnd: characterStream 

	"Private - Move cursor end of current line."
	| string |
	self closeTypeIn: characterStream.
	string := paragraph text string.
	self
		moveCursor:
			[:position | self walkAlongDisplayedLine
				ifTrue:[| targetLine |
					targetLine := paragraph lines at:(paragraph lineIndexOfCharacterIndex: position).
					targetLine = paragraph lines last
						ifTrue:[targetLine last + 1]
						ifFalse:[targetLine last]]
				ifFalse:[
					string
						indexOfAnyOf: CharacterSet crlf
						startingAt: position
						ifAbsent:[string size + 1]]]
		forward: true
		specialBlock:[:dummy | string size + 1].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 12/11/2009 13:38'!
cursorHome: characterStream 

	"Private - Move cursor from position in current line to beginning of
	current line. If control key is pressed put cursor at beginning of text"

	| string |

	string := paragraph text string.
	self
		moveCursor: [ :position | self walkAlongDisplayedLine
				ifTrue:[
					(paragraph lines at:(paragraph lineIndexOfCharacterIndex: position)) first]
				ifFalse:[
					(string
						lastIndexOfAnyOf: CharacterSet crlf
						startingAt: position - 1
						ifAbsent:[0]) + 1]]
		forward: false
		specialBlock: [:dummy | 1].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 9/19/2002 20:07'!
cursorLeft: characterStream 
	"Private - Move cursor left one character if nothing selected, otherwise 
	move cursor to beginning of selection. If the shift key is down, start 
	selecting or extending current selection. Don't allow cursor past 
	beginning of text"

	self closeTypeIn: characterStream.
	self
		moveCursor:[:position | position - 1 max: 1]
		forward: false
		specialBlock:[:position | self previousWord: position].
	^ true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 11/18/2002 17:09'!
cursorPageDown: characterStream 

	self closeTypeIn: characterStream.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine:[:lineNo | lineNo + self pageHeight]
				forward: true]
		forward: true
		specialBlock:[:dummy | dummy].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 11/18/2002 17:09'!
cursorPageUp: characterStream 

	self closeTypeIn: characterStream.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine:[:lineNo | lineNo - self pageHeight]
				forward: false]
		forward: false
		specialBlock:[:dummy | dummy].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 9/19/2002 20:01'!
cursorRight: characterStream 
	"Private - Move cursor right one character if nothing selected, 
	otherwise move cursor to end of selection. If the shift key is down, 
	start selecting characters or extending already selected characters. 
	Don't allow cursor past end of text"

	self closeTypeIn: characterStream.
	self
		moveCursor: [:position | position + 1]
		forward: true
		specialBlock:[:position | self nextWord: position].
	^ true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:32'!
cursorTopHome: characterStream 
	"Put cursor at beginning of text -- invoked from cmd-H shortcut, useful for keyboards that have no home key."
	
	self selectAt: 1.
	^ true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'th 11/18/2002 17:15'!
cursorUp: characterStream 

"Private - Move cursor from position in current line to same position in
prior line. If prior line too short, put at end"

	self closeTypeIn: characterStream.
	self
		moveCursor: [:position | self
				sameColumn: position
				newLine:[:line | line - 1]
				forward: false]
		forward: false
		specialBlock:[:dummy | dummy].
	^true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sw 8/1/97 16:33'!
cut
	"Cut out the current selection and redisplay the paragraph if necessary.  Undoer & Redoer: undoCutCopy:"

	self lineSelectAndEmptyCheck: [^ self].

	self replaceSelectionWith: self nullText. 
	self undoer: #undoCutCopy: with: self clipboardText.
	self clipboardTextPut: UndoSelection! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:33'!
cut: characterStream 
	"Cut out the current text selection.  Flushes typeahead."

	self cut.
	^true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 3/23/2009 17:48'!
debugIt: characterStream 
	"Triggered by Cmd-D; browse the thing represented by the current selection, if plausible"

	self debugIt.
	^ true! !

!ParagraphEditor methodsFor: 'current selection' stamp: ''!
deselect
	"If the text selection is visible on the screen, reverse its highlight."

	selectionShowing ifTrue: [self reverseSelection]! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'michael.rueger 2/23/2009 13:24'!
dispatchOnEnterWith: typeAheadStream
	"Enter key hit.  Treat is as an 'accept', viz a synonym for cmd-s.  If cmd key is down, treat is as a synonym for print-it. "

	self terminateAndInitializeAround: [
	Sensor commandKeyPressed
		ifTrue:
			[self printIt.]
		ifFalse: 
			[self closeTypeIn: typeAheadStream.
			self accept].
	].
	^ true! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'AlainPlantec 12/19/2009 23:52'!
dispatchOnKeyEvent: keyEvent with: typeAheadStream
	"Carry out the action associated with this character, if any.
	Type-ahead is passed so some routines can flush or use it."

	| honorCommandKeys keyValue keyChar char action |
	self clearParens.
	keyValue := OSPlatform current virtualKey: keyEvent scanCode.
  	keyValue := keyValue ifNil: [keyEvent keyValue].
	"Work around bug in some VMs delivering negative key values"
	keyChar := (keyValue max: 0) asCharacter.
	char := keyEvent keyCharacter.
	
	"mikki 1/3/2005 21:31 Preference for auto-indent on return added."
	keyChar = Character cr ifTrue: [
		^(self autoIndent
			xor: keyEvent controlKeyPressed)
			ifTrue: [self crWithIndent: typeAheadStream]
			ifFalse: [self normalCharacter: typeAheadStream character: char]].

	((honorCommandKeys := self cmdKeysInText)
		and: [keyChar = Character enter])
		ifTrue: [^ self dispatchOnEnterWith: typeAheadStream].
		
	(0 <= keyValue and: [keyValue < 256]) ifTrue: [	"none of the following is safe if the character's asciiValue is out of the 0..255 range"
		
	"Special keys overwrite crtl+key combinations - at least on Windows. To resolve this
	conflict, assume that keys other than cursor keys aren't used together with Crtl." 
	((self class specialShiftCmdKeys includes: keyValue) and: [keyValue < 27])
		ifTrue: [
			action := keyEvent controlKeyPressed
				ifTrue: [ShiftCmdActions at: keyValue + 1]
				ifFalse: [CmdActions at: keyValue + 1].
			^action numArgs = 1
				ifTrue: [self perform: action with: typeAheadStream]
				ifFalse: [self perform: action with: keyEvent with: typeAheadStream]].

	"backspace, and escape keys (ascii 8 and 27) are command keys"
	((honorCommandKeys and: [keyEvent commandKeyPressed]) or: [self class specialShiftCmdKeys includes: keyValue])
		ifTrue: [
			action := keyEvent leftShiftDown
					ifTrue: [ShiftCmdActions at: keyValue + 1]
					ifFalse: [CmdActions at: keyValue + 1].
			^action numArgs = 1
				ifTrue: [self perform: action with: typeAheadStream]
				ifFalse: [self perform: action with: typeAheadStream with: keyEvent]].

	"the control key can be used to invoke shift-cmd shortcuts"
	(honorCommandKeys and: [keyEvent controlKeyPressed])
		ifTrue: [
			action := ShiftCmdActions at: keyValue + 1.
			^action numArgs = 1
				ifTrue: [self perform: action with: typeAheadStream]
				ifFalse: [self perform: action with: typeAheadStream with: keyEvent]].
	
	"allow cut/copy/paste/selectAll regardless of cmdKeysInText preference.
	Useful when running a deployed/locked-down image (after disableProgrammerFacilities)."
	((#(cut: copySelection: paste: selectAll:) includes: (CmdActions at: keyValue + 1)) 
		and: [keyEvent commandKeyPressed])
		ifTrue: [
			action := CmdActions at: keyValue + 1.
			^action numArgs = 1
				ifTrue: [self perform: action with: typeAheadStream]
				ifFalse: [self perform: action with: typeAheadStream with: keyEvent]].
	]. "end of range protection"

	(')]}' includes: char)
		ifTrue: [self blinkPrevParen: char].

	^self normalCharacter: typeAheadStream character: char! !

!ParagraphEditor methodsFor: 'displaying' stamp: 'GuillermoPolito 5/24/2010 14:31'!
display
	"Redisplay the paragraph."

	| selectionState view |
	selectionState := selectionShowing.
	self deselect.
	paragraph foregroundColor: view foregroundColor
			backgroundColor: view backgroundColor;
			displayOn: Display.
	selectionState ifTrue: [self select]! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:33'!
displayIfFalse: characterStream 
	"Replace the current text selection with the text 'ifFalse:'--initiated by 
	ctrl-f."

	characterStream nextPutAll: 'ifFalse:'.
	^false! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:33'!
displayIfTrue: characterStream 
	"Replace the current text selection with the text 'ifTrue:'--initiated by 
	ctrl-t."

	characterStream nextPutAll: 'ifTrue:'.
	^false! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:33'!
doAgainMany: characterStream 
	"Do the previous thing again repeatedly. 1/26/96 sw"

	self closeTypeIn: characterStream.
	self againOrSame: (UndoMessage sends: #undoAgain:andReselect:typedKey:) many: true.
	^ true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:33'!
doAgainOnce: characterStream 
	"Do the previous thing again once. 1/26/96 sw"

	self closeTypeIn: characterStream.
	self again.
	^ true! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'di 5/10/1998 21:38'!
doIt
	"Set the context to include pool vars of the model.  Then evaluate."
	^ self evaluateSelection.
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:33'!
doIt: characterStream 
	"Called when user hits cmd-d.  Select the current line, if relevant, then evaluate and execute.  2/1/96 sw.
	2/29/96 sw: don't call selectLine; it's done by doIt now"

	self terminateAndInitializeAround: [self doIt].
	^ true! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'di 6/14/1998 13:08'!
doneTyping
	beginTypeInBlock := nil! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:33'!
duplicate: characterStream
	"Paste the current selection over the prior selection, if it is non-overlapping and
	 legal.  Flushes typeahead.  Undoer & Redoer: undoAndReselect."

	self closeTypeIn.
	(self hasSelection and: [self isDisjointFrom: otherInterval])
		ifTrue: "Something to duplicate"
			[self replace: otherInterval with: self selection and:
				[self selectAt: self pointIndex]]
		ifFalse:
			[self flash].
	^true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 13:25'!
enclose: characterStream keyEvent: keyEvent
	"Insert or remove bracket characters around the current selection.
	 Flushes typeahead."

	| char left right startIndex stopIndex oldSelection which text |
	char := keyEvent keyCharacter.
	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: char ifAbsent: [ ^true ].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := paragraph text.
	((startIndex > 1 and: [stopIndex <= text size])
		and:
		[(text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue:
			["already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse:
			["not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string ,(String with: right)
					emphasis: emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'nice 11/16/2009 14:41'!
encompassLine: anInterval
	"Return an interval that encompasses the entire line"
	| string left right |
	string := paragraph text string.
	left := (string lastIndexOfAnyOf: CharacterSet crlf startingAt: anInterval first - 1 ifAbsent:[0]) + 1.
	right := (string indexOfAnyOf: CharacterSet crlf startingAt: anInterval last + 1 ifAbsent: [string size + 1]) - 1.
	^left to: right! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 2/17/2010 01:53'!
escapeToDesktop: characterStream 
	"Pop up a morph to field keyboard input in the context of the desktop"

	ActiveWorld invokeWorldMenuFromEscapeKey.
	^ true! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'NikoSchwarz 6/5/2010 13:23'!
evaluateSelection
	"Treat the current selection as an expression; evaluate it and return the result"
	| result rcvr ctxt |
	self lineSelectAndEmptyCheck: [^ ''].

	(model respondsTo: #doItReceiver) 
		ifTrue: [FakeClassPool adopt: model selectedClass.  "Include model pool vars if any"
				rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: self selection asString
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [FakeClassPool adopt: nil. ^ #failedDoit]
			logged: true.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	FakeClassPool adopt: nil.
	^ result! !

!ParagraphEditor methodsFor: 'menu messages' stamp: ''!
exchange
	"See comment in exchangeWith:"

	self exchangeWith: otherInterval! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:34'!
exchange: characterStream
	"Exchange the current and prior selections.  Keeps typeahead."

	self closeTypeIn: characterStream.
	self exchange.
	^true! !

!ParagraphEditor methodsFor: 'private' stamp: 'alain.plantec 6/11/2008 15:06'!
exchangeWith: prior
	"If the prior selection is non-overlapping and legal, exchange the text of
	 it with the current selection and leave the currently selected text selected
	 in the location of the prior selection (or leave a caret after a non-caret if it was
	 exchanged with a caret).  If both selections are carets, flash & do nothing.
	 Don't affect the paste buffer.  Undoer: itself; Redoer: Undoer."

	| start stop before selection priorSelection delta altInterval |
	start := self startIndex.
	stop := self stopIndex - 1.
	((prior first <= prior last) | (start <= stop) "Something to exchange" and:
			[self isDisjointFrom: prior])
		ifTrue:
			[before := prior last < start.
			selection := self selection.
			priorSelection := paragraph text copyFrom: prior first to: prior last.

			delta := before ifTrue: [0] ifFalse: [priorSelection size - selection size].
			self zapSelectionWith: priorSelection.
			self selectFrom: prior first + delta to: prior last + delta.

			delta := before ifTrue: [stop - prior last] ifFalse: [start - prior first].
			self zapSelectionWith: selection.
			altInterval := prior first + delta to: prior last + delta.
			self undoer: #exchangeWith: with: altInterval.
			"If one was a caret, make it otherInterval & leave the caret after the other"
			prior first > prior last ifTrue: [self selectAt: UndoInterval last + 1].
			otherInterval := start > stop
				ifTrue: [self selectAt: altInterval last + 1. UndoInterval]
				ifFalse: [altInterval]]
		ifFalse:
			[self flash]! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sma 5/28/2000 09:34'!
experimentalCommand
	"Use for experimental command-key implementation.  Using this, 
	you can try things out without forever needing to reinitialize the 
	ParagraphEditor."

	self prettyPrint.
	^ true! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'alain.plantec 6/11/2008 15:08'!
exploreIt
	| result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
			ifTrue: [self flash]
			ifFalse: [result explore].
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:34'!
exploreIt: characterStream 
	"Explore the selection -- invoked via cmd-shift-I.  If there is no current selection, use the current line."

	self terminateAndInitializeAround: [self exploreIt].
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'di 9/7/1999 08:41'!
fileItIn
	"Make a Stream on the text selection and fileIn it.
	 1/24/96 sw: moved here from FileController; this function can be useful from any text window that shows stuff in chunk format"

	| selection |
	selection := self selection.
	self terminateAndInitializeAround:
		[(ReadWriteStream on: selection string from: 1 to: selection size) fileIn].
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:34'!
fileItIn: characterStream 
	"File in the selection; invoked via a keyboard shortcut, -- for now, cmd-shift-G."

	self terminateAndInitializeAround: [self fileItIn].
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'DamienCassou 9/29/2009 13:05'!
find
	"Prompt the user for a string to search for, and search the receiver from the current selection onward for it.  1/26/96 sw"

	| reply |
	reply := UIManager default request: 'Find what? ' translated initialAnswer: ''.
	reply isEmptyOrNil ifTrue: [^ self].
	self setSearch: reply.
	ChangeText := FindText.  "Implies no replacement to againOnce: method"
	self againOrSame: true
	
! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:34'!
find: characterStream
	"Prompt the user for what to find, then find it, searching from the current selection onward.  1/24/96 sw"

	self closeTypeIn: characterStream.
	self find.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: ''!
findAgain
	"Find the text-to-find again.  1/24/96 sw"

	self againOrSame: true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:34'!
findAgain: characterStream 
	"Find the desired text again.  1/24/96 sw"

	self closeTypeIn: characterStream.
	self findAgain.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: ''!
fit
	"Make the bounding rectangle of the paragraph contain all the text while 
	 not changing the width of the view of the paragraph.  No effect on undoability
	 of the preceding command."

	paragraph clearVisibleRectangle.
	paragraph fit.
	paragraph displayOn: Display; outline.
	self recomputeInterval! !

!ParagraphEditor methodsFor: 'displaying' stamp: ''!
flash
	"Causes the view of the paragraph to complement twice in succession."

	paragraph flash! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 13:29'!
forwardDelete: characterStream keyEvent: keyEvent
	"Delete forward over the next character.
	  Make Undo work on the whole type-in, not just the one char.
	wod 11/3/1998: If there was a selection use #zapSelectionWith: rather than #backspace: which was 'one off' in deleting the selection. Handling of things like undo or typeIn area were not fully considered."
	| startIndex usel upara uinterval ind stopIndex |
	startIndex := self mark.
	startIndex > paragraph text size
		ifTrue: [^ false].
	self hasSelection ifTrue:
		["there was a selection"
		self zapSelectionWith: self nullText.
		^ false].
	"Null selection - do the delete forward"
	beginTypeInBlock == nil	"no previous typing.  openTypeIn"
		ifTrue: [self openTypeIn. UndoSelection := self nullText].
	uinterval := UndoInterval deepCopy.
	upara := UndoParagraph deepCopy.
	stopIndex := startIndex.
	(keyEvent keyValue = 127 and: [keyEvent leftShiftDown])
		ifTrue: [stopIndex := (self nextWord: stopIndex) - 1].
	self selectFrom: startIndex to: stopIndex.
	self replaceSelectionWith: self nullText.
	self selectFrom: startIndex to: startIndex-1.
	UndoParagraph := upara.  UndoInterval := uinterval.
	UndoMessage selector == #noUndoer ifTrue: [
		(UndoSelection isText) ifTrue: [
			usel := UndoSelection.
			ind := startIndex. "UndoInterval startIndex"
			usel replaceFrom: usel size + 1 to: usel size with:
				(UndoParagraph text copyFrom: ind to: ind).
			UndoParagraph text replaceFrom: ind to: ind with:
self nullText]].
	^false! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/19/2002 18:22'!
hasCaret
	^self markBlock = self pointBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/19/2002 18:22'!
hasSelection
	^self hasCaret not! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'th 9/18/2002 16:31'!
hiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden info.  Return the entire string that was used by Cmd-6 to create this text attribute.  Usually enclosed in < >."

	| attrList |
	attrList := paragraph text attributesAt: (self pointIndex +
self markIndex)//2 forStyle: paragraph textStyle.
	attrList do: [:attr |
		(attr isKindOf: TextAction) ifTrue:
			[^ self selection asString, '<', attr info, '>']].
	"If none of the above"
	attrList do: [:attr |
		attr class == TextColor ifTrue:
			[^ self selection asString, '<', attr color printString, '>']].
	^ self selection asString, '[No hidden info]'! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 12:58'!
implementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ self flash].
	self terminateAndInitializeAround: 
		[ (self systemNavigation allImplementorsOf: aSelector) isEmpty
				ifTrue: [self internalCallToBrowseIt]
				ifFalse: [ self systemNavigation browseAllImplementorsOf: aSelector]]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'sd 5/24/2010 22:40'!
implementorsOfIt: characterStream 
	"Triggered by Cmd-m; browse implementors of the selector represented by the current selection, if plausible."

	self implementorsOfIt.
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'nice 11/30/2009 22:33'!
inOutdent: characterStream delta: delta
	"Add/remove a tab at the front of every line occupied by the selection. Flushes typeahead.  Derived from work by Larry Tesler back in December 1985.  Now triggered by Cmd-L and Cmd-R.  2/29/96 sw"

	| realStart realStop lines startLine stopLine start stop adjustStart indentation size numLines inStream newString outStream |

	"Operate on entire lines, but remember the real selection for re-highlighting later"
	realStart := self startIndex.
	realStop := self stopIndex - 1.

	"Special case a caret on a line of its own, including weird case at end of paragraph"
	(realStart > realStop and:
				[realStart < 2 or: [(paragraph string at: realStart - 1) == Character cr or: [(paragraph string at: realStart - 1) == Character lf]]])
		ifTrue:
			[delta < 0
				ifTrue:
					[self flash]
				ifFalse:
					[self replaceSelectionWith: Character tab asSymbol asText.
					self selectAt: realStart + 1].
			^ true].

	lines := paragraph lines.
	startLine := paragraph lineIndexOfCharacterIndex: realStart.
	stopLine := paragraph lineIndexOfCharacterIndex: (realStart max: realStop).
	start := (lines at: startLine) first.
	stop := (lines at: stopLine) last.
	
	"Pin the start of highlighting unless the selection starts a line"
	adjustStart := realStart > start.

	"Find the indentation of the least-indented non-blank line; never outdent more"
	indentation := (startLine to: stopLine) inject: 1000 into:
		[:m :l |
		 m min: (paragraph indentationOfLineIndex: l ifBlank: [:tabs | 1000])].			

	size :=  stop + 1 - start.
	numLines := stopLine + 1 - startLine.
	inStream := ReadStream on: paragraph string from: start to: stop.

	newString := WideString new: size + ((numLines * delta) max: 0).
	outStream := ReadWriteStream on: newString.

	"This subroutine does the actual work"
	self indent: delta fromStream: inStream toStream: outStream.

	"Adjust the range that will be highlighted later"
	adjustStart ifTrue: [realStart := (realStart + delta) max: start].
	realStop := realStop + outStream position - size.

	"Prepare for another iteration"
	indentation := indentation + delta.
	size := outStream position.
	inStream := outStream setFrom: 1 to: size.

	outStream == nil
		ifTrue: 	"tried to outdent but some line(s) were already left flush"
			[self flash]
		ifFalse:
			[self selectInvisiblyFrom: start to: stop.
			size = newString size ifFalse: [newString := outStream contents].
			self replaceSelectionWith: newString asText].
	self selectFrom: realStart to: realStop. 	"highlight only the original range"
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: ''!
indent: characterStream
	"Add a tab at the front of every line occupied by the selection. Flushes typeahead.  Invoked from keyboard via cmd-shift-R.  2/29/96 sw"

	^ self inOutdent: characterStream delta: 1! !

!ParagraphEditor methodsFor: 'private' stamp: ''!
indent: delta fromStream: inStream toStream: outStream
	"Append the contents of inStream to outStream, adding or deleting delta or -delta
	 tabs at the beginning, and after every CR except a final CR.  Do not add tabs
	 to totally empty lines, and be sure nothing but tabs are removed from lines."

	| ch skip cr tab prev atEnd |
	cr := Character cr.
	tab := Character tab.
	delta > 0
		ifTrue: "shift right"
			[prev := cr.
			 [ch := (atEnd := inStream atEnd) ifTrue: [cr] ifFalse: [inStream next].
			  (prev == cr and: [ch ~~ cr]) ifTrue:
				[delta timesRepeat: [outStream nextPut: tab]].
			  atEnd]
				whileFalse:
					[outStream nextPut: ch.
					prev := ch]]
		ifFalse: "shift left"
			[skip := delta. "a negative number"
			 [inStream atEnd] whileFalse:
				[((ch := inStream next) == tab and: [skip < 0]) ifFalse:
					[outStream nextPut: ch].
				skip := ch == cr ifTrue: [delta] ifFalse: [skip + 1]]]! !

!ParagraphEditor methodsFor: 'accessing' stamp: 'tk 4/21/1998 09:55'!
initialText
	^ initialText! !

!ParagraphEditor methodsFor: 'current selection' stamp: ''!
initializeSelection
	"Do the initial activity when starting up the receiver. For example, in the 
	ParagraphEditor highlight the current selection."

	self select! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'di 12/17/1998 09:41'!
insertAndSelect: aString at: anInteger

	self replace: (anInteger to: anInteger - 1)
		with: (Text string: (' ' , aString)
					attributes: emphasisHere)
		and: [self selectAndScroll]! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'th 9/17/2002 16:23'!
insertTypeAhead: typeAhead
	typeAhead position = 0 ifFalse:
		[self zapSelectionWith: (Text string: typeAhead contents emphasis: emphasisHere).
		typeAhead reset.
		self unselect]! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'alain.plantec 6/11/2008 15:09'!
inspectIt
	"1/13/96 sw: minor fixup"
	| result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
			ifTrue: [self flash]
			ifFalse: [result inspect].
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
inspectIt: characterStream 
	"Inspect the selection -- invoked via cmd-i.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine; it's done by inspectIt now"

self terminateAndInitializeAround: [self inspectIt].
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 12:58'!
internalCallToBrowseIt
	"Launch a browser for the class indicated by the current selection. 
	If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	self lineSelectAndEmptyCheck: [^ self].
	aClass := SystemNavigation default
				classFromPattern: (self selection string copyWithoutAll: CharacterSet crlf)
				withCaption: 'choose a class to browse...'.
	aClass ifNil: [^ self].
	self terminateAndInitializeAround: [ | aBrow |
			aBrow := SystemBrowser default new.
			aBrow setClass: aClass selector: nil.
			aBrow class
				openBrowserView: (aBrow openEditString: nil)
				label: 'System Browser']! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 12:57'!
internalCallToImplementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ self flash].
	self terminateAndInitializeAround: 
		[ (self systemNavigation allImplementorsOf: aSelector) isEmpty
				ifFalse: [ self systemNavigation browseAllImplementorsOf: aSelector]]! !

!ParagraphEditor methodsFor: 'private' stamp: 'cmm 4/9/2004 14:00'!
isDisjointFrom: anInterval
	"Answer true if anInterval is a caret not touching or within the current
	 interval, or if anInterval is a non-caret that does not overlap the current
	 selection."

	| fudge |
	fudge := anInterval size = 0 ifTrue: [1] ifFalse: [0].
	^(anInterval last + fudge < self startIndex or:
			[anInterval first - fudge >= self stopIndex])
! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
isDoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^(self isUndoing | self isRedoing) not! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
isRedoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^UndoParagraph == #redoing! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
isUndoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^UndoParagraph == #undoing! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'PeterHugossonMiller 9/3/2009 10:14'!
keystroke: keyEvent
	"Key struck on the keyboard. Find out which one and, if special, carry 
	out the associated special action. Otherwise, add the character to the 
	stream of characters.  Undoer & Redoer: see closeTypeIn."

	| typeAhead |
	typeAhead := (String new: 128) writeStream.
	self deselect.
	(self dispatchOnKeyEvent: keyEvent with: typeAhead)
		ifTrue: [
			self doneTyping.
			self setEmphasisHere.
			^self selectAndScroll; updateMarker].
	self openTypeIn.

	self hasSelection ifTrue: "save highlighted characters"
		[UndoSelection := self selection]. 
	self zapSelectionWith: 
			(Text string: typeAhead contents emphasis: emphasisHere).
	typeAhead reset.
	self unselect.
	self selectAndScroll.
	self updateMarker! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'th 9/19/2002 18:12'!
lineSelectAndEmptyCheck: returnBlock
	"If the current selection is an insertion point, expand it to be the entire current line; if after that's done the selection is still empty, then evaluate the returnBlock, which will typically consist of '[^ self]' in the caller -- check senders of this method to understand this."

	self selectLine.  "if current selection is an insertion point, then first select the entire line in which occurs before proceeding"
	self hasSelection ifFalse: [self flash.  ^ returnBlock value]! !

!ParagraphEditor methodsFor: 'private' stamp: 'nice 11/16/2009 10:11'!
lines
	"Other than my member paragraph i compute lines based on logical
	line breaks, not optical (which may change due to line wrapping of the editor)"
	| lines string index lineIndex stringSize |
	string := paragraph text string.
	"Empty strings have no lines at all. Think of something."
	string isEmpty ifTrue:[^{#(1 0 0)}].
	stringSize := string size.
	lines := OrderedCollection new: (string size // 15).
	index := 0.
	lineIndex := 0.
	string linesDo:[:line |
		lines addLast: (Array
			with: (index := index + 1)
			with: (lineIndex := lineIndex + 1)
			with: (index := index + line size min: stringSize))].
	"Special workaround for last line empty."
	(string last == Character cr or: [string last == Character lf])
	"lines last last < stringSize" ifTrue:[lines addLast:{stringSize +1. lineIndex+1. stringSize}].
	^lines! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
makeCapitalized: characterStream 
	"Force the current selection to be capitalized. Triggered by Cmd-Z."
	| prev |
prev := $-.  "not a letter"
	self replaceSelectionWith: (Text fromString:
			(self selection string collect:
				[:c | prev := prev isLetter ifTrue: [c asLowercase] ifFalse: [c asUppercase]])).
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
makeLowercase: characterStream 
	"Force the current selection to lowercase.  Triggered by Cmd-X."

	self replaceSelectionWith: (Text fromString: (self selection string asLowercase)).
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
makeUppercase: characterStream 
	"Force the current selection to uppercase.  Triggered by Cmd-Y."

	self replaceSelectionWith: (Text fromString: (self selection string asUppercase)).
	^ true! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 16:13'!
mark
	^ self markBlock stringIndex! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 10/21/2003 15:49'!
markBlock
	^ stopBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 10/21/2003 15:49'!
markBlock: aCharacterBlock
	stopBlock := aCharacterBlock.
! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/18/2002 12:31'!
markIndex
	^ self markBlock stringIndex! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 4/16/2003 19:31'!
methodNamesContainingIt
	"Open a browser on methods names containing the selected string"

	self lineSelectAndEmptyCheck: [^ self].
	Cursor wait showWhile:
		[self terminateAndInitializeAround: [self systemNavigation browseMethodsWhoseNamesContain: self selection string withBlanksTrimmed]].
	Cursor normal show! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
methodNamesContainingIt: characterStream 
	"Browse methods whose selectors containing the selection in their names"

	self methodNamesContainingIt.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'md 9/6/2005 18:45'!
methodSourceContainingIt
	"Open a browser on methods which contain the current selection in their source (case-sensitive full-text search of source). Slow!!"

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation browseMethodsWithSourceString: self selection string! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:37'!
methodStringsContainingIt: characterStream 
	"Invoked from cmd-E -- open a browser on all methods holding string constants containing it.  Flushes typeahead. "

	self methodStringsContainingit.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 4/16/2003 19:28'!
methodStringsContainingit
	"Open a browser on methods which contain the current selection as part of a string constant."

	self lineSelectAndEmptyCheck: [^ self].
	self terminateAndInitializeAround: [self systemNavigation browseMethodsWithString: self selection string]! !

!ParagraphEditor methodsFor: 'model access' stamp: 'michael.rueger 2/23/2009 18:31'!
model
	"Answer the receiver's model which is the same as the model of the 
	receiver's view."

	^model! !

!ParagraphEditor methodsFor: 'model access' stamp: 'michael.rueger 2/23/2009 18:31'!
model: aModel 
	"Controller|model: and Controller|view: are sent by View|controller: in 
	order to coordinate the links between the model, view, and controller. In 
	ordinary usage, the receiver is created and passed as the parameter to 
	View|controller: so that the receiver's model and view links can be set 
	up by the view."

	model := aModel! !

!ParagraphEditor methodsFor: 'private' stamp: 'michael.rueger 2/23/2009 13:30'!
moveCursor: directionBlock forward: forward specialBlock: specialBlock
	"Private - Move cursor.
	directionBlock is a one argument Block that computes the new Position from a given one.
	specialBlock is a one argumentBlock that computes the new position from a given one under the alternate semantics.
	Note that directionBlock always is evaluated first."
	| shift indices newPosition |
	shift := Sensor leftShiftDown.
	indices := self setIndices: shift forward: forward.
	newPosition := directionBlock value: (indices at: #moving).
	(Sensor commandKeyPressed or:[Sensor controlKeyPressed])
		ifTrue: [newPosition := specialBlock value: newPosition].
	shift
		ifTrue: [self selectMark: (indices at: #fixed) point: newPosition - 1]
		ifFalse: [self selectAt: newPosition]! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'di 5/6/1998 15:25'!
nextTokenFrom: start direction: dir
	"simple token-finder for compiler automated corrections"
	| loc str |
	loc := start + dir.
	str := paragraph text string.
	[(loc between: 1 and: str size) and: [(str at: loc) isSeparator]]
		whileTrue: [loc := loc + dir].
	^ loc! !

!ParagraphEditor methodsFor: 'private' stamp: 'sma 12/15/1999 11:32'!
nextWord: position
	| string index |
	string := paragraph text string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index + 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index + 1].
	^ index! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
noUndoer
	"The Undoer to use when the command can not be undone.  Checked for
	 specially by readKeyboard."

	UndoMessage := Message selector: #noUndoer! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:40'!
noop: characterStream 
	"Unimplemented keyboard command; just ignore it."

	^ true
! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 13:20'!
normalCharacter: characterStream character: character
	"A nonspecial character is to be added to the stream of characters."

	characterStream nextPut: character.
	^false! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'di 5/20/1998 08:31'!
notify: aString at: anInteger in: aStream 
	"The compilation of text failed. The syntax error is noted as the argument, 
	aString. Insert it in the text at starting character position anInteger."

	self insertAndSelect: aString at: (anInteger max: 1)! !

!ParagraphEditor methodsFor: 'private' stamp: ''!
nullText

	^Text string: '' emphasis: emphasisHere! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'nice 1/5/2010 15:59'!
objectsReferencingIt
	"Open a list inspector on all objects that reference the object that results when the current selection is evaluated.  "
	
	self terminateAndInitializeAround: [ | result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
		ifTrue: [self flash]
		ifFalse: [self systemNavigation
					browseAllObjectReferencesTo: result
					except: #()
					ifNone: [:obj | self flash]].
	]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'alain.plantec 2/6/2009 17:17'!
offerFontMenu
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.  
	Use only names of Fonts of this paragraph  "

	| aList reply |
	aList := paragraph textStyle fontNamesWithPointSizes.
	reply := UIManager default chooseFrom: aList values: aList.
	reply ~~ nil ifTrue:
		[self replaceSelectionWith:
			(Text string: self selection asString 
				attribute: (TextFontChange fontNumber: (aList indexOf: reply)))] ! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:40'!
offerFontMenu: characterStream 
	"The user typed the command key that requests a font change; Offer the font menu.  5/27/96 sw
	 Keeps typeahead.  (?? should flush?)"

	self closeTypeIn: characterStream.
	self offerFontMenu.
	^ true! !

!ParagraphEditor methodsFor: 'as yet unclassified' stamp: 'michael.rueger 2/23/2009 12:41'!
offerMenuFromEsc: aStream
	self yellowButtonActivity.
	^true "tell the caller that the character was processed "! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'th 9/18/2002 16:48'!
openTypeIn
	"Set up UndoSelection to null text (to be added to by readKeyboard and backTo:),
	 beginTypeInBlock to keep track of the leftmost backspace, and UndoParameter to tally
	 how many deleted characters were backspaced over rather than 'cut'.
	 You can't undo typing until after closeTypeIn."

	beginTypeInBlock == nil ifTrue:
		[UndoSelection := self nullText.
		self undoer: #noUndoer with: 0.
		beginTypeInBlock := self startIndex]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: ''!
outdent: characterStream
	"Remove a tab from the front of every line occupied by the selection. Flushes typeahead.  Invoked from keyboard via cmd-shift-L.  2/29/96 sw"

	^ self inOutdent: characterStream delta: -1! !

!ParagraphEditor methodsFor: 'private' stamp: 'th 9/20/2002 11:09'!
pageHeight
	| howManyLines visibleHeight totalHeight ratio |
	howManyLines := paragraph numberOfLines.
	visibleHeight := self visibleHeight.
	totalHeight := self totalTextHeight.
	ratio := visibleHeight / totalHeight.
	^(ratio * howManyLines) rounded - 2! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'th 9/20/2002 11:21'!
paste
	"Paste the text from the shared buffer over the current selection and 
	redisplay if necessary.  Undoer & Redoer: undoAndReselect."

	self replace: self selectionInterval with: self clipboardText and:
		[self selectAt: self pointIndex]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:41'!
paste: characterStream 
	"Replace the current text selection by the text in the shared buffer.
	 Keeps typeahead."

	self closeTypeIn: characterStream.
	self paste.
	^true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'JB 2/12/2010 15:22'!
pasteInitials: characterStream 
	"Replace the current text selection by an authorship name/date stamp; invoked by cmd-shift-v, easy way to put an authorship stamp in the comments of an editor.
	 Keeps typeahead."

	self closeTypeIn: characterStream.
	self replace: self selectionInterval with: (Text fromString: Author changeStamp) and: [self selectAt: self stopIndex].
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'ar 1/15/2001 18:36'!
pasteRecent
	"Paste an item chose from RecentClippings."

	| clipping |
	(clipping := Clipboard chooseRecentClipping) ifNil: [^ self].
	Clipboard clipboardText: clipping.
	^ self paste! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 10/21/2003 15:49'!
pointBlock
	^ startBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 10/21/2003 15:49'!
pointBlock: aCharacterBlock
	startBlock := aCharacterBlock.
! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/18/2002 12:31'!
pointIndex
	^ self pointBlock stringIndex! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 5/18/2009 15:58'!
prettyPrint
	"Reformat the contents of the receiver's view (a Browser)."

	| selectedClass newText |
	model selectedMessageCategoryName ifNil: [^ self flash].
	selectedClass := model selectedClassOrMetaClass.
	newText := selectedClass prettyPrinterClass
		format: self text
		in: selectedClass
		notifying: self.
	newText ifNotNil:
		[self deselect; selectInvisiblyFrom: 1 to: paragraph text size.
		self replaceSelectionWith: (newText asText makeSelectorBoldIn: selectedClass).
		self selectAt: 1]! !

!ParagraphEditor methodsFor: 'private' stamp: 'sma 12/15/1999 11:33'!
previousWord: position
	| string index |
	string := paragraph text string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index - 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index - 1].
	^ index + 1! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'AdrianKuhn 12/22/2009 02:53'!
printIt
	"Treat the current text selection as an expression; evaluate it. Insert the 
	description of the result of evaluation after the selection and then make 
	this description the new text selection."
	| result |
	result := self evaluateSelection.
	((result class == FakeClassPool) or: [result == #failedDoit])
			ifTrue: [self flash]
			ifFalse: [self afterSelectionInsertAndSelect: result printString]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:41'!
printIt: characterStream 
	"Print the results of evaluting the selection -- invoked via cmd-p.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine now, since it's called by doIt"

	self terminateAndInitializeAround: [self printIt].
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'dew 3/7/2000 21:06'!
printerSetup
	
	TextPrinter defaultTextPrinter inspect
! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:41'!
querySymbol: characterStream
	"Invoked by Ctrl-q to query the Symbol table and display alternate symbols.
	 See comment in completeSymbol:lastOffering: for details."

	self closeTypeIn: characterStream.	"keep typeahead"
	self hasCaret
		ifTrue: "Ctrl-q typed when a caret"
			[self perform: #completeSymbol:lastOffering: withArguments:
				((UndoParagraph == paragraph and: [UndoMessage sends: #undoQuery:lastOffering:])
					ifTrue: [UndoMessage arguments] "repeated Ctrl-q"
					ifFalse: [Array with: nil with: nil])] "initial Ctrl-q"
		ifFalse: "Ctrl-q typed when statements were highlighted"
			[self flash].
	^true! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'dvf 12/8/2001 00:46'!
raiseContextMenu: characterStream 
	"AFAIK, this is never called in morphic, because a subclass overrides it. Which is good, because a ParagraphEditor doesn't know about Morphic and thus duplicates the text-editing actions that really belong in the specific application, not the controller. So the context menu this would raise is likely to be out of date."
	self yellowButtonActivity.
	^true! !

!ParagraphEditor methodsFor: 'current selection' stamp: 'th 9/20/2002 11:41'!
recomputeInterval
	"The same characters are selected but their coordinates may have changed."

	self computeIntervalFrom: self mark to: self pointIndex - 1! !

!ParagraphEditor methodsFor: 'current selection' stamp: ''!
recomputeSelection
	"Redetermine the selection according to the start and stop block indices; 
	do not highlight."

	self deselect; recomputeInterval! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'lr 3/14/2010 21:13'!
referencesToIt
	"Open a references browser on the selected symbol"

	| aSymbol |
	self selectLine.
	((aSymbol := self selectedSymbol) == nil or: [ (Smalltalk globals includesKey: aSymbol) not ])
		ifTrue: [ ^ self flash ].
	self
		terminateAndInitializeAround: [ self systemNavigation browseAllCallsOn: (Smalltalk globals associationAt: self selectedSymbol) ]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:41'!
referencesToIt: characterStream 
	"Triggered by Cmd-N; browse references to the current selection"

	self referencesToIt.
	^ true! !

!ParagraphEditor methodsFor: 'accessing' stamp: ''!
replace: oldInterval with: newText and: selectingBlock 
	"Replace the text in oldInterval with newText and execute selectingBlock to establish the new selection.  Create an undoAndReselect:redoAndReselect: undoer to allow perfect undoing."

	| undoInterval |
	undoInterval := self selectionInterval.
	undoInterval = oldInterval ifFalse: [self selectInterval: oldInterval].
	UndoSelection := self selection.
	self zapSelectionWith: newText.
	selectingBlock value.
	otherInterval := self selectionInterval.
	self undoer: #undoAndReselect:redoAndReselect: with: undoInterval with: otherInterval! !

!ParagraphEditor methodsFor: 'accessing' stamp: ''!
replaceSelectionWith: aText
	"Remember the selection text in UndoSelection.
	 Deselect, and replace the selection text by aText.
	 Remember the resulting selectionInterval in UndoInterval and PriorInterval.
	 Set up undo to use UndoReplace."

	beginTypeInBlock ~~ nil ifTrue: [^self zapSelectionWith: aText]. "called from old code"
	UndoSelection := self selection.
	self zapSelectionWith: aText.
	self undoer: #undoReplace! !

!ParagraphEditor methodsFor: 'initialize-release' stamp: 'th 10/21/2003 15:49'!
resetState 
	"Establish the initial conditions for editing the paragraph: place caret 
	before first character, set the emphasis to that of the first character,
	and save the paragraph for purposes of canceling."

	stopBlock := paragraph defaultCharacterBlock.
	self pointBlock: stopBlock copy.
	beginTypeInBlock := nil.
	UndoInterval := otherInterval := 1 to: 0.
	self setEmphasisHere.
	selectionShowing := false.
	initialText := paragraph text copy! !

!ParagraphEditor methodsFor: 'current selection' stamp: 'BG 12/12/2003 12:50'!
reverseSelection
	"Reverse the valence of the current selection highlighting."
	selectionShowing := selectionShowing not.
	paragraph reverseFrom: self pointBlock to: self markBlock! !

!ParagraphEditor methodsFor: 'private' stamp: 'AlainPlantec 12/11/2009 13:39'!
sameColumn: start newLine: lineBlock forward: isForward
	"Private - Compute the index in my text
	with the line number derived from lineBlock,"
	" a one argument block accepting the old line number.
	The position inside the line will be preserved as good as possible"
	"The boolean isForward is used in the border case to determine if
	we should move to the beginning or the end of the line."
	| column currentLine offsetAtTargetLine targetEOL lines numberOfLines currentLineNumber targetLineNumber |
	self walkAlongDisplayedLine
		ifTrue: [
			lines := paragraph lines.
			numberOfLines := paragraph numberOfLines.
			currentLineNumber  := paragraph lineIndexOfCharacterIndex: start.
			currentLine := lines at: currentLineNumber]
		ifFalse: [
			lines := self lines.
			numberOfLines := lines size.
			currentLine := lines
				detect:[:lineInterval | lineInterval last >= start]
				ifNone:[lines last].
			currentLineNumber := currentLine second].
	column := start - currentLine first.
	targetLineNumber := ((lineBlock value: currentLineNumber) max: 1) min: numberOfLines.
	offsetAtTargetLine := (lines at: targetLineNumber) first.
	targetEOL := (lines at: targetLineNumber) last + (targetLineNumber == numberOfLines ifTrue:[1]ifFalse:[0]).
	targetLineNumber == currentLineNumber
	"No movement or movement failed. Move to beginning or end of line."
		ifTrue:[^isForward
			ifTrue:[targetEOL]
			ifFalse:[offsetAtTargetLine]].
	^offsetAtTargetLine + column min: targetEOL.! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:41'!
save: characterStream
	"Submit the current text.  Equivalent to 'accept' 1/18/96 sw
	 Keeps typeahead."

	self closeTypeIn: characterStream.
	self terminateAndInitializeAround: [self accept].
	^ true! !

!ParagraphEditor methodsFor: 'scrolling' stamp: 'BG 12/12/2003 15:31'!
scrollBy: heightToMove
	"Move the paragraph by heightToMove, and reset the text selection."
	^ paragraph scrollBy: heightToMove withSelectionFrom: self pointBlock to: self markBlock! !

!ParagraphEditor methodsFor: 'scrolling' stamp: ''!
scrollToBottom
	"Scroll so that the tail end of the text is visible in the view.  5/6/96 sw"

	self scrollView: (paragraph clippingRectangle bottom 
		- paragraph compositionRectangle bottom)! !

!ParagraphEditor methodsFor: 'scrolling' stamp: ''!
scrollToTop
	"Scroll so that the paragraph is at the top of the view."

	self scrollView: (paragraph clippingRectangle top 
		- paragraph compositionRectangle top)! !

!ParagraphEditor methodsFor: 'scrolling' stamp: ''!
scrollView: anInteger 
	"Paragraph scrolling uses opposite polarity"
	^ self scrollBy: anInteger negated! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:41'!
search: characterStream
	"Invoked by Ctrl-S.  Same as 'again', but always uses the existing FindText
	 and ChangeText regardless of the last edit."

	self closeTypeIn: characterStream.
	self againOrSame: true. "true means use same keys"
	^true! !

!ParagraphEditor methodsFor: 'current selection' stamp: ''!
select
	"If the text selection is visible on the screen, highlight it."

	selectionShowing ifFalse: [self reverseSelection]! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'sw 8/29/2000 14:58'!
selectAll
	"Make the selection be all the characters of the receiver"

	self selectFrom: 1 to: paragraph text string size! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'michael.rueger 2/23/2009 12:41'!
selectAll: characterStream 
	"select everything, invoked by cmd-a.  1/17/96 sw"

	self closeTypeIn: characterStream.
	self selectFrom: 1 to: paragraph text string size.
	^ true! !

!ParagraphEditor methodsFor: 'current selection' stamp: 'StephaneDucasse 5/27/2010 22:15'!
selectAndScroll
	"Scroll until the selection is in the view and then highlight it."
	| lineHeight deltaY clippingRectangle endBlock |
	self select.
	endBlock := self stopBlock.
	lineHeight := paragraph textStyle lineGrid.
	clippingRectangle := paragraph clippingRectangle.
	deltaY := endBlock top - clippingRectangle top.
	deltaY >= 0 
		ifTrue: [deltaY := endBlock bottom - clippingRectangle bottom max: 0].
		"check if stopIndex below bottom of clippingRectangle"
	deltaY ~= 0 
		ifTrue: [self scrollBy: (deltaY abs + lineHeight - 1 truncateTo: lineHeight) * deltaY sign]! !

!ParagraphEditor methodsFor: 'current selection' stamp: 'StephaneDucasse 5/27/2010 22:15'!
selectAndScrollToTop
	"Scroll until the selection is in the view and then highlight it."
	| lineHeight deltaY clippingRectangle |
	self select.
	lineHeight := paragraph textStyle lineGrid.
	clippingRectangle := paragraph clippingRectangle.
	deltaY := self stopBlock top - clippingRectangle top.
	deltaY ~= 0 
		ifTrue: [self scrollBy: (deltaY abs + lineHeight - 1 truncateTo: lineHeight) * deltaY sign]! !

!ParagraphEditor methodsFor: 'new selection' stamp: ''!
selectAt: characterIndex 
	"Deselect, then place the caret before the character at characterIndex.
	 Be sure it is in view."

	self selectFrom: characterIndex to: characterIndex - 1! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'michael.rueger 2/23/2009 12:42'!
selectCurrentTypeIn: characterStream 
	"Select what would be replaced by an undo (e.g., the last typeIn)."

	| prior |

	self closeTypeIn: characterStream.
	prior := otherInterval.
	self closeTypeIn: characterStream.
	self selectInterval: UndoInterval.
	otherInterval := prior.
	^ true! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/18/2002 16:50'!
selectFrom: start to: stop
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	(start = self startIndex and: [stop + 1 = self stopIndex]) ifFalse:
		[self deselect.
		self selectInvisiblyFrom: start to: stop].
	self selectAndScroll! !

!ParagraphEditor methodsFor: 'new selection' stamp: ''!
selectInterval: anInterval
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	self selectFrom: anInterval first to: anInterval last! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'di 5/9/1998 20:59'!
selectInvisiblyFrom: start to: stop
	"Select the designated characters, inclusive.  Make no visual changes."

	^ self computeIntervalFrom: start to: stop! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/18/2002 14:17'!
selectInvisiblyMark: mark point: point
	"Select the designated characters, inclusive.  Make no visual changes."

	^ self computeIntervalFrom: mark to: point! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/19/2002 17:17'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	self hasSelection ifTrue:[^self].
	self selectInterval: (self encompassLine: self selectionInterval)! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/18/2002 14:18'!
selectMark: mark point: point
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	(mark =  self markIndex and: [point + 1 = self pointIndex]) ifFalse:
		[self deselect.
		self selectInvisiblyMark: mark point: point].
	self selectAndScroll! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/19/2002 18:49'!
selectPrecedingIdentifier
	"Invisibly select the identifier that ends at the end of the selection, if any."

	| string sep stop tok |
	tok := false.
	string := paragraph text string.
	stop := self stopIndex - 1.
	[stop > 0 and: [(string at: stop) isSeparator]] whileTrue: [stop := stop - 1].
	sep := stop.
	[sep > 0 and: [(string at: sep) tokenish]] whileTrue: [tok := true. sep := sep - 1].
	tok ifTrue: [self selectInvisiblyFrom: sep + 1 to: stop]! !

!ParagraphEditor methodsFor: 'new selection' stamp: 'th 9/18/2002 16:51'!
selectWord
	"Select delimited text or word--the result of double-clicking."

	| openDelimiter closeDelimiter direction match level leftDelimiters rightDelimiters
	string here hereChar start stop |
	string := paragraph text string.
	here := self pointIndex.
	(here between: 2 and: string size)
		ifFalse: ["if at beginning or end, select entire string"
			^self selectFrom: 1 to: string size].
	leftDelimiters := '([{<''"
'.
	rightDelimiters := ')]}>''"
'.
	openDelimiter := string at: here - 1.
	match := leftDelimiters indexOf: openDelimiter.
	match > 0
		ifTrue: 
			["delimiter is on left -- match to the right"
			start := here.
			direction := 1.
			here := here - 1.
			closeDelimiter := rightDelimiters at: match]
		ifFalse: 
			[openDelimiter := string at: here.
			match := rightDelimiters indexOf: openDelimiter.
			match > 0
				ifTrue: 
					["delimiter is on right -- match to the left"
					stop := here - 1.
					direction := -1.
					closeDelimiter := leftDelimiters at: match]
				ifFalse: ["no delimiters -- select a token"
					direction := -1]].
	level := 1.
	[level > 0 and: [direction > 0
			ifTrue: [here < string size]
			ifFalse: [here > 1]]]
		whileTrue: 
			[hereChar := string at: (here := here + direction).
			match = 0
				ifTrue: ["token scan goes left, then right"
					hereChar tokenish
						ifTrue: [here = 1
								ifTrue: 
									[start := 1.
									"go right if hit string start"
									direction := 1]]
						ifFalse: [direction < 0
								ifTrue: 
									[start := here + 1.
									"go right if hit non-token"
									direction := 1]
								ifFalse: [level := 0]]]
				ifFalse: ["bracket match just counts nesting level"
					hereChar = closeDelimiter
						ifTrue: [level := level - 1"leaving nest"]
						ifFalse: [hereChar = openDelimiter 
									ifTrue: [level := level + 1"entering deeper nest"]]]].

	level > 0 ifTrue: ["in case ran off string end"	here := here + direction].
	direction > 0
		ifTrue: [self selectFrom: start to: here - 1]
		ifFalse: [self selectFrom: here + 1 to: stop]! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'michael.rueger 2/23/2009 12:42'!
selectWord: characterStream
	self closeTypeIn: characterStream.
	self selectWord.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'sd 5/24/2010 22:41'!
selectedSelector
	"Try to make a selector out of the current text selection"
	
	^self selection string findSelector! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'nice 11/16/2009 11:33'!
selectedSymbol
	"Return the currently selected symbol, or nil if none.  Spaces, tabs and returns are ignored"

	| aString |
	self hasCaret ifTrue: [^ nil].
	aString := self selection string.
	aString isOctetString ifTrue: [aString := aString asOctetString].
	aString := aString copyWithoutAll:
		{Character space.  Character cr.  Character lf. Character tab}.
	aString size == 0 ifTrue: [^ nil].
	Symbol hasInterned: aString  ifTrue: [:sym | ^ sym].

	^ nil! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'yo 7/31/2004 16:27'!
selection
	"Answer the text in the paragraph that is currently selected."

	| t |
	t := paragraph text copyFrom: self startIndex to: self stopIndex - 1.
	t string isOctetString ifTrue: [t asOctetStringText].
	^ t.
! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/19/2002 18:10'!
selectionAsStream
	"Answer a ReadStream on the text in the paragraph that is currently 
	selected."

	^ReadWriteStream
		on: paragraph string
		from: self startIndex
		to: self stopIndex - 1! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/18/2002 16:18'!
selectionInterval
	"Answer the interval that is currently selected."

	^self startIndex to: self stopIndex - 1 ! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 6/11/2008 15:13'!
sendersOfIt
	"Open a senders browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ self flash].
	self terminateAndInitializeAround: [self systemNavigation browseAllCallsOn: aSelector]! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:42'!
sendersOfIt: characterStream 
	"Triggered by Cmd-n; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self sendersOfIt.
	^ true! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'th 9/18/2002 17:28'!
setAlignment: aSymbol
	| attr interval |
	attr := TextAlignment perform: aSymbol.
	interval := self encompassLine: self selectionInterval.
	paragraph replaceFrom: interval first to: interval last with:
		((paragraph text copyFrom: interval first to: interval last) addAttribute: attr) displaying: true.
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'yo 5/27/2004 13:56'!
setEmphasis: emphasisSymbol
	"Change the emphasis of the current selection."

	| oldAttributes attribute |
	oldAttributes := paragraph text attributesAt: self selectionInterval first forStyle: paragraph textStyle.

	attribute := TextEmphasis perform: emphasisSymbol.
	(emphasisSymbol == #normal) 
		ifFalse:	[oldAttributes do:	
			[:att | (att dominates: attribute) ifTrue: [attribute turnOff]]].
	self replaceSelectionWith: (self selection addAttribute: attribute)! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'th 9/18/2002 16:49'!
setEmphasisHere

	emphasisHere := (paragraph text attributesAt: (self pointIndex - 1 max: 1) forStyle: paragraph textStyle)
					select: [:att | att mayBeExtended]! !

!ParagraphEditor methodsFor: 'private' stamp: 'md 2/22/2006 21:17'!
setIndices: shiftPressed forward: forward
	"Little helper method that sets the moving and fixed indices according to some flags."
	| indices |
	indices := Dictionary new.
	shiftPressed ifTrue: [
			indices at: #moving put: self pointIndex.
			indices at: #fixed put: self markIndex
		] ifFalse: [
			forward
				ifTrue:[
					indices at: #moving put: self stopIndex.
					indices at: #fixed put: self startIndex.
				] ifFalse: [
					indices at: #moving put: self startIndex.
					indices at: #fixed put: self stopIndex.
				]
		].
	^indices! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 13:02'!
setMark: anIndex
	self markBlock: (paragraph characterBlockForIndex: anIndex)
! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 13:02'!
setPoint: anIndex
	self pointBlock: (paragraph characterBlockForIndex: anIndex)
! !

!ParagraphEditor methodsFor: 'accessing' stamp: ''!
setSearch: aString
	"Set the FindText and ChangeText to seek aString; except if already seeking aString, leave ChangeText alone so again will repeat last replacement."

	FindText string = aString
		ifFalse: [FindText := ChangeText := aString asText]! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'alain.plantec 6/11/2008 15:13'!
setSearchString
	"Make the current selection, if any, be the current search string."
	self hasCaret ifTrue: [self flash. ^ self].
	self setSearch:  self selection string! !

!ParagraphEditor methodsFor: 'nonediting/nontyping keys' stamp: 'michael.rueger 2/23/2009 12:42'!
setSearchString: characterStream
	"Establish the current selection as the current search string."

	| aString |
	self closeTypeIn: characterStream.
	self lineSelectAndEmptyCheck: [^ true].
	aString :=  self selection string.
	aString size == 0
		ifTrue:
			[self flash]
		ifFalse:
			[self setSearch: aString].
	^ true! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 13:22'!
shiftEnclose: characterStream keyEvent: keyEvent
	"Insert or remove bracket characters around the current selection.
	 Flushes typeahead."

	| char left right startIndex stopIndex oldSelection which text |
	char := keyEvent keyCharacter.
	char = $9 ifTrue: [ char := $( ].
	char = $, ifTrue: [ char := $< ].
	char = $[ ifTrue: [ char := ${ ].
	char = $' ifTrue: [ char := $" ].
	char asciiValue = 27 ifTrue: [ char := ${ ].	"ctrl-["

	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: char ifAbsent: [1].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := paragraph text.
	((startIndex > 1 and: [stopIndex <= text size])
		and:
		[(text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue:
			["already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse:
			["not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string ,(String with: right)
					emphasis: emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!ParagraphEditor methodsFor: 'typing/selecting keys' stamp: 'th 9/19/2002 17:34'!
simulatedBackspace
	"Backspace over the last character, derived from hand-char recognition.  2/5/96 sw"

	| startIndex |
	startIndex := self markIndex + (self hasSelection ifTrue: [1] ifFalse: [0]).

	startIndex := 1 max: startIndex - 1.
	self backTo: startIndex.
	^ false! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'alain.plantec 6/11/2008 15:11'!
simulatedKeystroke: char
	"Accept char as if it were struck on the keyboard.  This version does not yet deal with command keys, and achieves update in the receiver's typically inactive window via the sledge-hammer of uncache-bits."

	self deselect.
	self openTypeIn.
	self markBlock = self pointBlock ifFalse: [UndoSelection := self selection].
	self zapSelectionWith:
		(Text string: char asString emphasis: emphasisHere).
	self userHasEdited.
	self unselect.
	self selectAndScroll.
	self updateMarker.
! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 16:10'!
startBlock
	^ self pointBlock min: self markBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 13:10'!
startBlock: aCharacterBlock
	self markBlock: aCharacterBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/18/2002 14:27'!
startIndex
	^ self startBlock stringIndex! !

!ParagraphEditor methodsFor: 'typing support' stamp: 'di 10/6/1998 08:45'!
startOfTyping
	"Compatibility during change from characterBlock to integer"
	beginTypeInBlock == nil ifTrue: [^ nil].
	beginTypeInBlock isNumber ifTrue: [^ beginTypeInBlock].
	"Last line for compatibility during change from CharacterBlock to Integer."
	^ beginTypeInBlock stringIndex
	! !

!ParagraphEditor methodsFor: 'initialize-release' stamp: 'di 5/15/2000 13:51'!
stateArray
	^ {ChangeText.
		FindText.
		UndoInterval.
		UndoMessage.
		UndoParagraph.
		UndoSelection.
		Undone.
		self selectionInterval.
		self startOfTyping.
		emphasisHere}! !

!ParagraphEditor methodsFor: 'initialize-release' stamp: 'di 10/5/1998 17:03'!
stateArrayPut: stateArray
	| sel |
	ChangeText := stateArray at: 1.
	FindText := stateArray at: 2.
	UndoInterval := stateArray at: 3.
	UndoMessage := stateArray at: 4.
	UndoParagraph := stateArray at: 5.
	UndoSelection := stateArray at: 6.
	Undone := stateArray at: 7.
	sel := stateArray at: 8.
	self selectFrom: sel first to: sel last.
	beginTypeInBlock := stateArray at: 9.
	emphasisHere := stateArray at: 10.! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 16:14'!
stopBlock
	^ self pointBlock max: self markBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 13:10'!
stopBlock: aCharacterBlock
	self pointBlock: aCharacterBlock! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/18/2002 14:27'!
stopIndex
	^ self stopBlock stringIndex! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:42'!
swapChars: characterStream 
	"Triggered byCmd-Y;.  Swap two characters, either those straddling the insertion point, or the two that comprise the selection.  Suggested by Ted Kaehler.  "

	| currentSelection aString chars |
	(chars := self selection) size == 0
		ifTrue:
			[currentSelection := self pointIndex.
			self selectMark: currentSelection - 1 point: currentSelection]
		ifFalse:
			[chars size == 2
				ifFalse:
					[self flash.  ^ true]
				ifTrue:
					[currentSelection := self pointIndex - 1]].
	aString := self selection string.
	self replaceSelectionWith: (Text string: aString reversed emphasis: emphasisHere).
	self selectAt: currentSelection + 1.
	^ true! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'ab 3/23/2005 16:49'!
tallyIt

	^ self tallySelection! !

!ParagraphEditor methodsFor: 'do-its' stamp: 'nice 1/5/2010 15:59'!
tallySelection
	"Treat the current selection as an expression; evaluate and tally it."
	| receiver context v |

	(model respondsTo: #doItReceiver) 
		ifTrue: 
			[FakeClassPool adopt: model selectedClass. "Include model pool vars if any"
			receiver := model doItReceiver.
			context := model doItContext]
		ifFalse:
			[receiver := context := nil].
	self lineSelectAndEmptyCheck: [ ^ self ].

	[ | compiledMethod |
		compiledMethod := self compileSelectionFor: receiver in: context.
		compiledMethod ifNil: [^ self].
		TimeProfileBrowser spyOn: [
			v := compiledMethod valueWithReceiver: receiver arguments: #()].
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	FakeClassPool adopt: nil.
	
	self inform: ('Result: ', (v printStringLimitedTo: 20)).
! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:43'!
tempCommand: characterStream 
	"Experimental.  Triggered by Cmd-t; put trial cmd-key commands here to see how they work, before hanging them on their own cmd accelerators."

	self experimentalCommand.
	^ true! !

!ParagraphEditor methodsFor: 'as yet unclassified' stamp: 'michael.rueger 2/23/2009 18:37'!
terminateAndInitializeAround: aBlock
	"1/12/96 sw"
	self controlTerminate.
	aBlock value.
	self controlInitialize! !

!ParagraphEditor methodsFor: 'accessing' stamp: ''!
text
	"Answer the text of the paragraph being edited."

	^paragraph text! !

!ParagraphEditor methodsFor: 'as yet unclassified' stamp: 'sbw 10/13/1999 22:40'!
totalTextHeight

	^paragraph boundingBox height! !

!ParagraphEditor methodsFor: 'menu messages' stamp: 'michael.rueger 2/23/2009 13:23'!
undo
	"Reset the state of the paragraph prior to the previous edit.
	 If another ParagraphEditor instance did that edit, UndoInterval is invalid;
	 just recover the contents of the undo-buffer at the start of the paragraph."

	Sensor flushKeyboard. "a way to flush stuck keys"
	self closeTypeIn.

	UndoParagraph == paragraph ifFalse: "Can't undo another paragraph's edit"
		[UndoMessage := Message selector: #undoReplace.
		UndoInterval := 1 to: 0.
		Undone := true].
	UndoInterval ~= self selectionInterval ifTrue: "blink the actual target"
		[self selectInterval: UndoInterval; deselect].

	"Leave a signal of which phase is in progress"
	UndoParagraph := Undone ifTrue: [#redoing] ifFalse: [#undoing].
	UndoMessage sentTo: self.
	UndoParagraph := paragraph! !

!ParagraphEditor methodsFor: 'editing keys' stamp: 'michael.rueger 2/23/2009 12:43'!
undo: characterStream 
	"Undo the last edit.  Keeps typeahead, so undo twice is a full redo."

	self closeTypeIn: characterStream.
	self undo.
	^true! !

!ParagraphEditor methodsFor: 'undoers' stamp: 'nice 1/5/2010 15:59'!
undoAgain: indices andReselect: home typedKey: wasTypedKey
	"The last command was again.  Undo it. Redoer: itself."

	| findSize substText |
	(self isRedoing & wasTypedKey) ifTrue: "redelete search key"
		[self selectInterval: home.
		self zapSelectionWith: self nullText].

	findSize := (self isRedoing ifTrue: [FindText] ifFalse: [ChangeText]) size.
	substText := self isUndoing ifTrue: [FindText] ifFalse: [ChangeText].
	(self isUndoing ifTrue: [indices size to: 1 by: -1] ifFalse: [1 to: indices size]) do:
		[:i | | index subject |
		index := indices at: i.
		(subject := index to: index + findSize - 1) = self selectionInterval ifFalse:
			[self selectInterval: subject].
		FindText == ChangeText ifFalse: [self zapSelectionWith: substText]].

	self isUndoing
		ifTrue:  "restore selection to where it was when 'again' was invoked"
			[wasTypedKey
				ifTrue: "search started by typing key at a caret; restore it"
					[self selectAt: home first.
					self zapSelectionWith: FindText.
					self selectAt: home last + 1]
				ifFalse: [self selectInterval: home]].

	self undoMessage: UndoMessage forRedo: self isUndoing! !

!ParagraphEditor methodsFor: 'undoers' stamp: ''!
undoAndReselect: undoHighlight redoAndReselect: redoHighlight
	"Undo typing, cancel, paste, and other operations that are like replaces
	 but the selection is not the whole restored text after undo, redo, or both.
	 undoHighlight is selected after this phase and redoHighlight after the next phase.
	Redoer: itself."

	self replace: self selectionInterval with: UndoSelection and:
		[self selectInterval: undoHighlight].
	self undoMessage: (UndoMessage argument: redoHighlight) forRedo: self isUndoing
! !

!ParagraphEditor methodsFor: 'undoers' stamp: ''!
undoCutCopy: oldPasteBuffer
	"Undo of a cut, copy, or any edit that changed CurrentSelection.  Be sure
	 undo-copy does not lock the model.  Redoer: itself, so never isRedoing."

	| recentCut |
	recentCut := self clipboardText.	
	UndoSelection size = UndoInterval size
		ifFalse: [self replaceSelectionWith: UndoSelection].
	self clipboardTextPut: oldPasteBuffer.
	self undoer: #undoCutCopy: with: recentCut! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
undoMessage: aMessage forRedo: aBoolean
	"Call this from an undoer/redoer to set up UndoMessage as the
	 corresponding redoer/undoer.  Also set up UndoParagraph, as well
	 as the state variable Undone.  It is assumed that UndoInterval has been
	 established (generally by zapSelectionWith:) and that UndoSelection has been
	 saved (generally by replaceSelectionWith: or replace:With:and:)."

	self isDoing ifTrue: [UndoParagraph := paragraph].
	UndoMessage := aMessage.
	Undone := aBoolean! !

!ParagraphEditor methodsFor: 'undoers' stamp: 'th 9/19/2002 18:46'!
undoQuery: hintText lastOffering: selectorOrNil
	"Undo ctrl-q.  selectorOrNil (if not nil) is the previously offered selector.
	 hintText is the original hint.  Redoer: completeSymbol."

	self zapSelectionWith: UndoSelection.
	self undoMessage: (Message selector: #completeSymbol:lastOffering: arguments: UndoMessage arguments) forRedo: true.
	self selectAt: self stopIndex! !

!ParagraphEditor methodsFor: 'undoers' stamp: ''!
undoReplace
	"Undo of any command that replaced a selection by other text that it left
	 highlighted, and that is undone and redone by simple reversal of the
	 operation.  This is the most common Undoer; call replaceSelectionWith:
	 to get this setup.  Redoer: itself, so never isRedoing."

	self replaceSelectionWith: UndoSelection! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
undoer: aSelector
	"See comment in undoMessage:.  Use this version when aSelector has no arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector) forRedo: false! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
undoer: aSelector with: arg1
	"See comment in undoMessage:.  Use this version when aSelector has one argument, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector argument: arg1) forRedo: false! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
undoer: aSelector with: arg1 with: arg2
	"See comment in undoMessage:.  Use this version when aSelector has two arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector arguments: (Array with: arg1 with: arg2)) forRedo: false! !

!ParagraphEditor methodsFor: 'undo support' stamp: ''!
undoer: aSelector with: arg1 with: arg2 with: arg3
	"See comment in undoMessage:.  Use this version when aSelector has three arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector arguments: (Array with: arg1 with: arg2 with: arg3)) forRedo: false! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/17/2002 16:23'!
unselect
	self markBlock: self pointBlock copy.! !

!ParagraphEditor methodsFor: 'scrolling' stamp: 'alain.plantec 6/11/2008 15:45'!
updateMarker
	"A variation of computeMarkerRegion--only redisplay the marker in the scrollbar if an actual change has occurred in the positioning of the paragraph."
! !

!ParagraphEditor methodsFor: 'accessing' stamp: 'jm 3/18/98 20:38'!
userHasEdited
	"Note that the user has edited my text. Here it is just a noop so that the Character Recognizer won't fail when used with a vanilla ParagrahEditor."
! !

!ParagraphEditor methodsFor: 'as yet unclassified' stamp: 'sbw 10/13/1999 22:33'!
visibleHeight

	^paragraph clippingRectangle height! !

!ParagraphEditor methodsFor: 'settings' stamp: 'AlainPlantec 12/11/2009 13:38'!
walkAlongDisplayedLine
	^ self class walkAlongDisplayedLine! !

!ParagraphEditor methodsFor: 'watchIt' stamp: 'hfm 1/11/2009 10:35'!
watchIt

	self inspectIt! !

!ParagraphEditor methodsFor: 'accessing-selection' stamp: 'th 9/19/2002 18:12'!
zapSelectionWith: aText
	"Deselect, and replace the selection text by aText.
	 Remember the resulting selectionInterval in UndoInterval and otherInterval.
	 Do not set up for undo."

	| start stop |
	self deselect.
	start := self startIndex.
	stop := self stopIndex.
	(aText isEmpty and: [stop > start]) ifTrue:
		["If deleting, then set emphasisHere from 1st character of the deletion"
		emphasisHere := (paragraph text attributesAt: start forStyle: paragraph textStyle)
					select: [:att | att mayBeExtended]].
	(start = stop and: [aText size = 0]) ifFalse:
		[paragraph
			replaceFrom: start
			to: stop - 1
			with: aText
			displaying: true.
		self computeIntervalFrom: start to: start + aText size - 1.
		UndoInterval := otherInterval := self selectionInterval]! !

DisplayObject subclass: #Path
	instanceVariableNames: 'form collectionOfPoints'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!Path commentStamp: '<historical>' prior: 0!
I am the abstract superclass of the Graphic spatial primitives. I represent an ordered sequence of Points. Spatial primitives are used to generate "trajectories" such as lines and circles.!

Path subclass: #Arc
	instanceVariableNames: 'quadrant radius center'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!Arc commentStamp: '<historical>' prior: 0!
Arcs are an unusual implementation of splines due to Ted Kaehler.  Imagine two lines that meet at a corner. Now imagine two moving points; one moves from the corner to the end on one line, the other moves from the end of the other line in to the corner.  Now imagine a series of lines drawn between those moving points at each step along the way (they form a sort of spider web pattern).  By connecting segments of the intersecting lines, a smooth curve is achieved that is tangent to both of the original lines.  Voila.!

!Arc class methodsFor: 'examples' stamp: ''!
example
	"Click the button somewhere on the screen. The designated point will
	be the center of an Arc with radius 50 in the 4th quadrant."

	| anArc aForm |
	aForm := Form extent: 1 @ 30.	"make a long thin Form for display"
	aForm fillBlack.						"turn it black"
	anArc := Arc new.
	anArc form: aForm.					"set the form for display"
	anArc radius: 50.0.
	anArc center: Sensor waitButton.
	anArc quadrant: 4.
	anArc displayOn: Display.
	Sensor waitButton

	"Arc example"! !

!Arc methodsFor: 'accessing' stamp: ''!
center
	"Answer the point at the center of the receiver."

	^center! !

!Arc methodsFor: 'accessing' stamp: ''!
center: aPoint 
	"Set aPoint to be the receiver's center."

	center := aPoint! !

!Arc methodsFor: 'accessing' stamp: ''!
center: aPoint radius: anInteger 
	"The receiver is defined by a point at the center and a radius. The 
	quadrant is not reset."

	center := aPoint.
	radius := anInteger! !

!Arc methodsFor: 'accessing' stamp: ''!
center: aPoint radius: anInteger quadrant: section 
	"Set the receiver's quadrant to be the argument, section. The size of the 
	receiver is defined by the center and its radius."

	center := aPoint.
	radius := anInteger.
	quadrant := section! !

!Arc methodsFor: 'display box access' stamp: ''!
computeBoundingBox
	| aRectangle aPoint |
	aRectangle := center - radius + form offset extent: form extent + (radius * 2) asPoint.
	aPoint := center + form extent.
	quadrant = 1 ifTrue: [^ aRectangle encompass: center x @ aPoint y].
	quadrant = 2 ifTrue: [^ aRectangle encompass: aPoint x @ aPoint y].
	quadrant = 3 ifTrue: [^ aRectangle encompass: aPoint x @ center y].
	quadrant = 4 ifTrue: [^ aRectangle encompass: center x @ center y]! !

!Arc methodsFor: 'displaying' stamp: 'nice 1/5/2010 15:59'!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger fillColor: aForm

	| nSegments line angle sin cos xn yn |
	nSegments := 12.0.
	line := Line new.
	line form: self form.
	angle := 90.0 / nSegments.
	sin := (angle * (2 * Float pi / 360.0)) sin.
	cos := (angle * (2 * Float pi / 360.0)) cos.
	quadrant = 1
		ifTrue: 
			[xn := radius asFloat.
			yn := 0.0].
	quadrant = 2
		ifTrue: 
			[xn := 0.0.
			yn := 0.0 - radius asFloat].
	quadrant = 3
		ifTrue: 
			[xn := 0.0 - radius asFloat.
			yn := 0.0].
	quadrant = 4
		ifTrue: 
			[xn := 0.0.
			yn := radius asFloat].
	nSegments asInteger
		timesRepeat: 
			[ | xn1 yn1 |xn1 := xn * cos + (yn * sin).
			yn1 := yn * cos - (xn * sin).
			line beginPoint: center + (xn asInteger @ yn asInteger).
			line endPoint: center + (xn1 asInteger @ yn1 asInteger).
			line
				displayOn: aDisplayMedium
				at: aPoint
				clippingBox: clipRect
				rule: anInteger
				fillColor: aForm.
			xn := xn1.
			yn := yn1]! !

!Arc methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: aTransformation clippingBox: clipRect rule: anInteger fillColor: aForm

	| newArc tempCenter |
	newArc := Arc new.
	tempCenter := aTransformation applyTo: self center.
	newArc center: tempCenter x asInteger @ tempCenter y asInteger.
	newArc quadrant: self quadrant.
	newArc radius: (self radius * aTransformation scale x) asInteger.
	newArc form: self form.
	newArc
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

!Arc methodsFor: 'accessing' stamp: ''!
quadrant
	"Answer the part of the circle represented by the receiver."
	^quadrant! !

!Arc methodsFor: 'accessing' stamp: ''!
quadrant: section 
	"Set the part of the circle represented by the receiver to be the argument, 
	section."

	quadrant := section! !

!Arc methodsFor: 'accessing' stamp: ''!
radius
	"Answer the receiver's radius."

	^radius! !

!Arc methodsFor: 'accessing' stamp: ''!
radius: anInteger 
	"Set the receiver's radius to be the argument, anInteger."

	radius := anInteger! !

Arc subclass: #Circle
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!Circle commentStamp: '<historical>' prior: 0!
I represent a full circle. I am made from four Arcs.!

!Circle class methodsFor: 'examples' stamp: ''!
exampleOne 
	"Click any button somewhere on the screen. The point will be the center
	of the circcle of radius 150."

	| aCircle aForm |
	aForm := Form extent: 1@30.
	aForm fillBlack.
	aCircle := Circle new.
	aCircle form: aForm.
	aCircle radius: 150.
	aCircle center: Sensor waitButton.
	aCircle displayOn: Display
	
	"Circle exampleOne"! !

!Circle class methodsFor: 'examples' stamp: ''!
exampleTwo
	"Designate a rectangular area that should be used as the brush for
	displaying the circle. Click any button at a point on the screen which
	will be the center location for the circle. The curve will be displayed
	with a long black form."

	| aCircle aForm |
	aForm := Form fromUser.
	aCircle := Circle new.
	aCircle form: aForm.
	aCircle radius: 150.
	aCircle center: Sensor waitButton.
	aCircle displayOn: Display at: 0 @ 0 rule: Form reverse
 
	 "Circle exampleTwo"! !

!Circle methodsFor: 'display box access' stamp: ''!
computeBoundingBox

	^center - radius + form offset extent: form extent + (radius * 2) asPoint! !

!Circle methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger fillColor: aForm

	1 to: 4 do:
		[:i |
		super quadrant: i.
		super displayOn: aDisplayMedium
			at: aPoint
			clippingBox: clipRect
			rule: anInteger
			fillColor: aForm]! !

!Circle methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: aTransformation clippingBox: clipRect rule: anInteger fillColor: aForm

	1 to: 4 do:
		[:i |
		super quadrant: i.
		super displayOn: aDisplayMedium
			transformation: aTransformation
			clippingBox: clipRect
			rule: anInteger
			fillColor: aForm]! !

Path subclass: #CurveFitter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!CurveFitter commentStamp: '<historical>' prior: 0!
I represent a conic section determined by three points p1, p2 and p3. I interpolate p1 and p3 and am tangent to line p1, p2 at p1 and line p3, p2 at p3.!

!CurveFitter class methodsFor: 'examples' stamp: '6/9/97 10:16 di'!
example
	"Designate three locations on the screen by clicking any button. The
	curve determined by the points will be displayed with a long black form."

	| aCurveFitter aForm |  
	aForm := Form extent: 1@30.			"make a long thin Form for display "
	aForm fillBlack.							"turn it black"
	aCurveFitter := CurveFitter new.
	aCurveFitter form: aForm.						"set the form for display"
				"collect three Points and show them on the dispaly"
	aCurveFitter firstPoint: Sensor waitButton. Sensor waitNoButton.
	aForm displayOn: Display at: aCurveFitter firstPoint.
	aCurveFitter secondPoint: Sensor waitButton. Sensor waitNoButton.
	aForm displayOn: Display at: aCurveFitter secondPoint.
	aCurveFitter thirdPoint: Sensor waitButton. Sensor waitNoButton.
	aForm displayOn: Display at: aCurveFitter thirdPoint.

	aCurveFitter displayOn: Display					"display the CurveFitter"

	"CurveFitter example"! !

!CurveFitter class methodsFor: 'instance creation' stamp: ''!
new

	| newSelf | 
	newSelf := super new: 3.
	newSelf add: 0@0.
	newSelf add: 0@0.
	newSelf add: 0@0.
	^newSelf! !

!CurveFitter methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger fillColor: aForm

	| pa pb k s p1 p2 p3 line |
	line := Line new.
	line form: self form.
	collectionOfPoints size < 3 ifTrue: [self error: 'Curve must have three points'].
	p1 := self firstPoint.
	p2 := self secondPoint.
	p3 := self thirdPoint.
	s := Path new.
	s add: p1.
	pa := p2 - p1.
	pb := p3 - p2.
	k := 5 max: pa x abs + pa y abs + pb x abs + pb y abs // 20.
	"k is a guess as to how many line segments to use to approximate 
	the curve."
	1 to: k do: 
		[:i | 
		s add: pa * i // k + p1 * (k - i) + (pb * (i - 1) // k + p2 * (i - 1)) // (k - 1)].
	s add: p3.
	1 to: s size - 1 do: 
		[:i | 
		line beginPoint: (s at: i).
		line endPoint: (s at: i + 1).
		line displayOn: aDisplayMedium
			at: aPoint
			clippingBox: clipRect
			rule: anInteger
			fillColor: aForm]! !

!CurveFitter methodsFor: 'displaying' stamp: '6/9/97 10:16 di'!
displayOn: aDisplayMedium transformation: aTransformation clippingBox: clipRect rule: anInteger fillColor: aForm

	| transformedPath newCurveFitter |
	transformedPath := aTransformation applyTo: self.
	newCurveFitter := CurveFitter new.
	newCurveFitter firstPoint: transformedPath firstPoint.
	newCurveFitter secondPoint: transformedPath secondPoint.
	newCurveFitter thirdPoint: transformedPath thirdPoint.
	newCurveFitter form: self form.
	newCurveFitter
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

Path subclass: #Line
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!Line commentStamp: '<historical>' prior: 0!
I represent the line segment specified by two points.!

!Line class methodsFor: 'examples' stamp: ''!
example
	"Designate two places on the screen by clicking any mouse button. A
	straight path with a square black form will be displayed connecting the
	two selected points."

	| aLine aForm |  
	aForm := Form extent: 20@20.		"make a form one quarter of inch square"
	aForm fillBlack.							"turn it black"
	aLine := Line new.
	aLine form: aForm.						"use the black form for display"
	aLine beginPoint: Sensor waitButton. Sensor waitNoButton.
	aForm displayOn: Display at: aLine beginPoint.	
	aLine endPoint: Sensor waitButton.
	aLine displayOn: Display.				"display the line"

	"Line example"! !

!Line class methodsFor: 'instance creation' stamp: ''!
from: beginPoint to: endPoint withForm: aForm 
	"Answer an instance of me with end points begingPoint and endPoint; 
	the source form for displaying the line is aForm."

	| newSelf | 
	newSelf := super new: 2.
	newSelf add: beginPoint.
	newSelf add: endPoint.
	newSelf form: aForm.
	^newSelf! !

!Line class methodsFor: 'instance creation' stamp: ''!
new

	| newSelf | 
	newSelf := super new: 2.
	newSelf add: 0@0.
	newSelf add: 0@0.
	^newSelf! !

!Line methodsFor: 'accessing' stamp: ''!
beginPoint
	"Answer the first end point of the receiver."

	^self first! !

!Line methodsFor: 'accessing' stamp: ''!
beginPoint: aPoint 
	"Set the first end point of the receiver to be the argument, aPoint. 
	Answer aPoint."

	self at: 1 put: aPoint.
	^aPoint! !

!Line methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger fillColor: aForm 
	"The form associated with this Path will be displayed, according  
	to one of the sixteen functions of two logical variables (rule), at  
	each point on the Line. Also the source form will be first anded  
	with aForm as a mask. Does not effect the state of the Path."

	collectionOfPoints size < 2 ifTrue: [self error: 'a line must have two points'].
	aDisplayMedium
		drawLine: self form
		from: self beginPoint + aPoint
		to: self endPoint + aPoint
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

!Line methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: aTransformation clippingBox: clipRect rule: anInteger fillColor: aForm

	| newPath newLine |
	newPath := aTransformation applyTo: self.
	newLine := Line new.
	newLine beginPoint: newPath firstPoint.
	newLine endPoint: newPath secondPoint.
	newLine form: self form.
	newLine
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

!Line methodsFor: 'displaying' stamp: ''!
displayOnPort: aPort at: aPoint 
	aPort sourceForm: self form; combinationRule: Form under; fillColor: nil.
	aPort drawFrom: collectionOfPoints first + aPoint
		to: collectionOfPoints last + aPoint! !

!Line methodsFor: 'accessing' stamp: ''!
endPoint
	"Answer the last end point of the receiver."

	^self last! !

!Line methodsFor: 'accessing' stamp: ''!
endPoint: aPoint 
	"Set the first end point of the receiver to be the argument, aPoint. 
	Answer aPoint."

	self at: 2 put: aPoint.
	^aPoint! !

Path subclass: #LinearFit
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!LinearFit commentStamp: '<historical>' prior: 0!
I represent a piece-wise linear approximation to a set of points in the plane.!

!LinearFit class methodsFor: 'examples' stamp: ''!
example
	"Select points on a Path using the red button. Terminate by selecting
	any other button. Creates a Path from the points and displays it as a
	piece-wise linear approximation." 

	| aLinearFit aForm flag |
	aLinearFit := LinearFit new.
	aForm := Form extent: 1 @ 40.
	aForm  fillBlack.
	aLinearFit form: aForm.
	flag := true.
	[flag] whileTrue:
		[Sensor waitButton.
		 Sensor redButtonPressed
			ifTrue: [aLinearFit add: Sensor waitButton. Sensor waitNoButton.
					aForm displayOn: Display at: aLinearFit last]
			ifFalse: [flag:=false]].
	aLinearFit displayOn: Display

	"LinearFit example"! !

!LinearFit methodsFor: 'displaying' stamp: 'jrm 9/7/1999 22:16'!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger
fillColor: aForm
 
	| line |
	line := Line new.
	line form: self form.
	1 to: self size - 1 do: 
		[:i | 
		line beginPoint: (self at: i).
		line endPoint: (self at: i + 1).
		line displayOn: aDisplayMedium
			at: aPoint
			clippingBox: clipRect
			rule: anInteger
			fillColor: aForm]! !

!LinearFit methodsFor: 'displaying' stamp: 'jrm 9/7/1999 23:00'!
displayOn: aDisplayMedium transformation: aTransformation clippingBox:
clipRect rule: anInteger fillColor: aForm 

	| transformedPath |
	"get the scaled and translated Path."
	transformedPath := aTransformation applyTo: self.
	transformedPath
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

!Path class methodsFor: 'examples' stamp: ''!
example
	"Creates a Path from mousePoints and displays it several ways on the display screen. Messes up the display. For learning about class Path, just select the code below and execute it to create a path and see it redisplayed in another place on the screen. Each path displays using a different form. A path is indicated by pressing the red mouse button in a sequence; press any other mouse button to terminate. "

	| aPath aForm pl fl flag |
	aForm := Form extent: 2 @ 40.		"creates a form one inch long"
	aForm fillBlack.							"turns it black"
	aPath := Path new.
	aPath form: aForm.						"use the long black form for displaying"
	flag := true.
	[flag]
		whileTrue: 
			[Sensor waitButton.
			Sensor redButtonPressed
				ifTrue: 
					[aPath add: Sensor waitButton.
					Sensor waitNoButton.
					aForm displayOn: Display at: aPath last]
				ifFalse: [flag := false]].
	Display fillWhite.
	aPath displayOn: Display.			"the original path"
	pl := aPath translateBy: 0 @ 100.
	fl := Form extent: 40 @ 40.
	fl fillGray.
	pl form: fl.
	pl displayOn: Display.				"the translated path"
	Sensor waitNoButton

	"Path example"! !

!Path class methodsFor: 'instance creation' stamp: ''!
new

	^self basicNew initializeCollectionOfPoints! !

!Path class methodsFor: 'instance creation' stamp: ''!
new: anInteger

	^self basicNew initializeCollectionOfPoints: anInteger! !

!Path methodsFor: 'adding' stamp: 'StephaneDucasse 5/6/2010 09:13'!
add: aPoint 
	"Include aPoint as one of the receiver's elements."

	^ collectionOfPoints add: aPoint! !

!Path methodsFor: 'accessing' stamp: ''!
at: index 
	"Answer the point on the receiver's path at position index."

	^collectionOfPoints at: index! !

!Path methodsFor: 'accessing' stamp: ''!
at: index put: aPoint 
	"Store the argument, aPoint, as the point on the receiver's path at position
	index."

	^collectionOfPoints at: index put: aPoint! !

!Path methodsFor: 'enumerating' stamp: ''!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect the resulting values into a path that is like the receiver. Answer 
	the new path."

	| newCollection |
	newCollection := collectionOfPoints collect: aBlock.
	newCollection form: self form.
	^newCollection! !

!Path methodsFor: 'display box access' stamp: ''!
computeBoundingBox 
	"Refer to the comment in DisplayObject|computeBoundingBox."

	| box |
	box := Rectangle origin: (self at: 1) extent: 0 @ 0.
	collectionOfPoints do: 
		[:aPoint | box := box merge: (Rectangle origin: aPoint extent: 0 @ 0)].
	^box! !

!Path methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm 
	"Display this Path--offset by aPoint, clipped by clipRect and the form 
	associated with this Path will be displayedr according to one of the sixteen 
	functions of two logical variables (rule). Also the source form will be first 
	anded with aForm as a mask. Does not effect the state of the Path"

	collectionOfPoints do: 
		[:element | 
		self form
			displayOn: aDisplayMedium
			at: element + aDisplayPoint
			clippingBox: clipRectangle
			rule: ruleInteger
			fillColor: aForm]! !

!Path methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle rule: ruleInteger fillColor: aForm 
	"Displays this path, translated and scaled by aTransformation. Get the
	scaled and translated Path."

	| newPath transformedPath |
	transformedPath := displayTransformation applyTo: self.
	newPath := Path new.
	transformedPath do: [:point | newPath add: point].
	newPath form: self form.
	newPath
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRectangle
		rule: ruleInteger
		fillColor: aForm! !

!Path methodsFor: 'accessing' stamp: ''!
first
	"Answer the first point on the receiver's path; included to correspond to 
	OrderedCollection protocol."

	^collectionOfPoints first! !

!Path methodsFor: 'accessing' stamp: ''!
firstPoint
	"Answer the first point on the receiver's path."

	^collectionOfPoints first! !

!Path methodsFor: 'accessing' stamp: ''!
firstPoint: aPoint 
	"Replace the first element of the receiver with the new value aPoint. 
	Answer the argument aPoint."

	collectionOfPoints at: 1 put: aPoint.
	^aPoint! !

!Path methodsFor: 'accessing' stamp: ''!
form
	"Answer the receiver's form, or, if form is nil, then answer a 1 x 1 black 
	form (a black dot)."

	| aForm |
	form == nil
		ifTrue: 
			[aForm := Form extent: 1 @ 1.
			aForm fillBlack.
			^aForm]
		ifFalse: 
			[^form]! !

!Path methodsFor: 'accessing' stamp: ''!
form: aForm 
	"Make the argument, aForm, be the receiver's form."

	form := aForm! !

!Path methodsFor: 'private' stamp: ''!
initializeCollectionOfPoints

	collectionOfPoints := OrderedCollection new! !

!Path methodsFor: 'private' stamp: ''!
initializeCollectionOfPoints: anInteger

	collectionOfPoints := OrderedCollection new: anInteger! !

!Path methodsFor: 'testing' stamp: ''!
isEmpty

	^collectionOfPoints isEmpty! !

!Path methodsFor: 'accessing' stamp: ''!
last
	"Answer the last point on the receiver's path; included to correspond to 
	OrderedCollection protocol."

	^collectionOfPoints last! !

!Path methodsFor: 'accessing' stamp: ''!
offset
	"There are basically two kinds of display objects in the system: those
	that, when asked to transform themselves, create a new object; and those
	that side effect themselves by maintaining a record of the transformation
	request (typically an offset). Path, like Rectangle and Point, is a display
	object of the first kind."

	self shouldNotImplement! !

!Path methodsFor: 'removing' stamp: 'di 4/4/2000 12:33'!
removeAllSuchThat: aBlock 
	"Evaluate aBlock for each element of the receiver.
	Remove each element for which aBlock evaluates to true."

	collectionOfPoints removeAllSuchThat: aBlock.
! !

!Path methodsFor: 'transforming' stamp: 'jrm 9/1/1999 21:26'!
scaleBy: aPoint 
	"Answers a new Path scaled by aPoint. Does not affect the current data in 
	this Path."

	| newPath | 
	newPath := self species new: self size. 
	newPath form: self form.
	collectionOfPoints do: [:element | newPath add: (element scaleBy: aPoint)].
	^newPath! !

!Path methodsFor: 'accessing' stamp: ''!
secondPoint
	"Answer the second element of the receiver."

	^collectionOfPoints at: 2! !

!Path methodsFor: 'accessing' stamp: ''!
secondPoint: aPoint 
	"Replace the second element of the receiver with the new value aPoint. 
	Answer the argument aPoint."

	collectionOfPoints at: 2 put: aPoint.
	^aPoint! !

!Path methodsFor: 'enumerating' stamp: ''!
select: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new path like the receiver only those elements for which 
	aBlock evaluates to true. Answer the new path."

	| newCollection |
	newCollection := collectionOfPoints select: aBlock.
	newCollection form: self form.
	^newCollection! !

!Path methodsFor: 'accessing' stamp: ''!
size
	"Answer the length of the receiver."

	^collectionOfPoints size! !

!Path methodsFor: 'accessing' stamp: ''!
thirdPoint
	"Answer the third element of the receiver."

	^collectionOfPoints at: 3! !

!Path methodsFor: 'accessing' stamp: ''!
thirdPoint: aPoint 
	"Replace the third element of the receiver with the new value aPoint. 
	Answer the argument aPoint."

	collectionOfPoints at: 3 put: aPoint.
	^aPoint! !

!Path methodsFor: 'transforming' stamp: 'jrm 9/1/1999 21:28'!
translateBy: aPoint 
	"Answers a new Path whose elements are translated by aPoint. Does not
	affect the elements of this Path."

	| newPath |
	newPath := self species new: self size.
	newPath form: self form.
	collectionOfPoints do: [:element | newPath add: (element translateBy: aPoint)].
	^newPath! !

Path subclass: #Spline
	instanceVariableNames: 'coefficients'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ST80-Paths'!

!Spline commentStamp: '<historical>' prior: 0!
I represent a collection of Points through which a cubic spline curve is fitted.!

!Spline class methodsFor: 'examples' stamp: '6/8/97 13:55 di'!
example
	"Designate points on the Path by clicking the red button. Terminate by
	pressing any other button. A curve will be displayed, through the
	selected points, using a long black form."

	| splineCurve aForm flag|
	aForm := Form extent: 2@2.
	aForm  fillBlack.
	splineCurve := Spline new.
	splineCurve form: aForm.
	flag := true.
	[flag] whileTrue:
		[Sensor waitButton.
		 Sensor redButtonPressed
			ifTrue: 
				[splineCurve add: Sensor waitButton. 
				 Sensor waitNoButton.
				 aForm displayOn: Display at: splineCurve last]
			ifFalse: [flag:=false]].
	splineCurve computeCurve.
	splineCurve isEmpty 
		ifFalse: [splineCurve displayOn: Display.
				Sensor waitNoButton].
 
	"Spline example"! !

!Spline methodsFor: 'accessing' stamp: ''!
coefficients
	"Answer an eight-element Array of Arrays each of which is the length 
	of the receiver. The first four arrays are the values, first, second and 
	third derivatives, respectively, for the parametric spline in x. The last 
	four elements are for y."

	^coefficients! !

!Spline methodsFor: 'displaying' stamp: ''!
computeCurve
	"Compute an array for the coefficients."

	| length extras |
	length := self size.
	extras := 0.
	coefficients := Array new: 8.
	1 to: 8 do: [:i | coefficients at: i put: (Array new: length + extras)].
	1 to: 5 by: 4 do: 
		[:k | 
		1 to: length do:
			[:i | (coefficients at: k)
					at: i put: (k = 1
						ifTrue: [(self at: i) x asFloat]
						ifFalse: [(self at: i) y asFloat])].
			1 to: extras do: [:i | (coefficients at: k)
					at: length + i put: ((coefficients at: k)
						at: i + 1)].
			self derivs: (coefficients at: k)
				first: (coefficients at: k + 1)
				second: (coefficients at: k + 2)
				third: (coefficients at: k + 3)].
	extras > 0 
		ifTrue: [1 to: 8 do: 
					[:i | 
					coefficients at: i put: ((coefficients at: i)
											copyFrom: 2 to: length + 1)]]! !

!Spline methodsFor: 'private' stamp: ''!
derivs: a first: point1 second: point2 third: point3
	"Compute the first, second and third derivitives (in coefficients) from
	the Points in this Path (coefficients at: 1 and coefficients at: 5)."

	| l v anArray |
	l := a size.
	l < 2 ifTrue: [^self].
	l > 2
	  ifTrue:
		[v := Array new: l.
		 v  at:  1 put: 4.0.
		 anArray := Array new: l.
		 anArray  at:  1 put: (6.0 * ((a  at:  1) - ((a  at:  2) * 2.0) + (a  at:  3))).
		 2 to: l - 2 do:
			[:i | 
			v  at:  i put: (4.0 - (1.0 / (v  at:  (i - 1)))).
			anArray
				at:  i 
				put: (6.0 * ((a  at:  i) - ((a  at:  (i + 1)) * 2.0) + (a  at:  (i + 2)))
						- ((anArray  at:  (i - 1)) / (v  at:  (i - 1))))].
		 point2  at: (l - 1) put: ((anArray  at:  (l - 2)) / (v  at:  (l - 2))).
		 l - 2 to: 2 by: 0-1 do: 
			[:i | 
			point2 
				at: i 
				put: ((anArray  at:  (i - 1)) - (point2  at:  (i + 1)) / (v  at:  (i - 1)))]].
	point2 at: 1 put: (point2  at:  l put: 0.0).
	1 to: l - 1 do:
		[:i | point1 
				at: i 
				put: ((a at: (i + 1)) - (a  at:  i) - 
						((point2  at:  i) * 2.0 + (point2  at:  (i + 1)) / 6.0)).
		      point3 at: i put: ((point2  at:  (i + 1)) - (point2  at:  i))]! !

!Spline methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger fillColor: aForm 
	"Display the receiver, a spline curve, approximated by straight line
	segments."

	| n line t x y x1 x2 x3 y1 y2 y3 |
	collectionOfPoints size < 1 ifTrue: [self error: 'a spline must have at least one point'].
	line := Line new.
	line form: self form.
	line beginPoint: 
		(x := (coefficients at: 1) at: 1) rounded @ (y := (coefficients at: 5) at: 1) rounded.
	1 to: (coefficients at: 1) size - 1 do: 
		[:i | 
		"taylor series coefficients"
		x1 := (coefficients at: 2) at: i.
		y1 := (coefficients at: 6) at: i.
		x2 := ((coefficients at: 3) at: i) / 2.0.
		y2 := ((coefficients at: 7) at: i) / 2.0.
		x3 := ((coefficients at: 4) at: i) / 6.0.
		y3 := ((coefficients at: 8) at: i) / 6.0.
		"guess n"
		n := 5 max: (x2 abs + y2 abs * 2.0 + ((coefficients at: 3)
							at: i + 1) abs + ((coefficients at: 7)
							at: i + 1) abs / 100.0) rounded.
		1 to: n - 1 do: 
			[:j | 
			t := j asFloat / n.
			line endPoint: 
				(x3 * t + x2 * t + x1 * t + x) rounded 
							@ (y3 * t + y2 * t + y1 * t + y) rounded.
			line
				displayOn: aDisplayMedium
				at: aPoint
				clippingBox: clipRect
				rule: anInteger
				fillColor: aForm.
			line beginPoint: line endPoint].
		line beginPoint: 
				(x := (coefficients at: 1) at: i + 1) rounded 
					@ (y := (coefficients at: 5) at: i + 1) rounded.
		line
			displayOn: aDisplayMedium
			at: aPoint
			clippingBox: clipRect
			rule: anInteger
			fillColor: aForm]! !

!Spline methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: aTransformation clippingBox: clipRect rule: anInteger fillColor: aForm 
	"Get the scaled and translated path of newKnots."

	| newKnots newSpline |
	newKnots := aTransformation applyTo: self.
	newSpline := Spline new.
	newKnots do: [:knot | newSpline add: knot].
	newSpline form: self form.
	newSpline
		displayOn: aDisplayMedium
		at: 0 @ 0
		clippingBox: clipRect
		rule: anInteger
		fillColor: aForm! !

DisplayText subclass: #Paragraph
	instanceVariableNames: 'clippingRectangle compositionRectangle destinationForm rule mask marginTabsLevel lines lastLine'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'ST80-Support'!

!Paragraph commentStamp: '<historical>' prior: 0!
I represent displayable text that has been decoraged with margin alignment, line leading, and tab settings.!

!Paragraph class methodsFor: 'examples' stamp: 'tk 9/30/96'!
example
	"This simple example illustrates how to display a few lines of text on the screen at the current cursor point.  
	Fixed. "

	| para point |
	point := Sensor waitButton.
	para := 'This is the first line of characters
and this is the second line.' asParagraph.
	para displayOn: Display at: point.

	"Paragraph example"! !

!Paragraph class methodsFor: 'instance creation' stamp: ''!
new
	"Do not allow an uninitialized view. Create with text that has no
	characters."

	^self withText: '' asText! !

!Paragraph class methodsFor: 'instance creation' stamp: ''!
withText: aText 
	"Answer an instance of me with text set to aText and style set to the 
	system's default text style."

	^self withText: aText style: DefaultTextStyle copy! !

!Paragraph class methodsFor: 'instance creation' stamp: ''!
withText: aText style: aTextStyle 
	"Answer an instance of me with text set to aText and style set to 
	aTextStyle."

	^super new setWithText: aText style: aTextStyle! !

!Paragraph class methodsFor: 'instance creation' stamp: ''!
withText: aText style: aTextStyle compositionRectangle: compRect clippingRectangle: clipRect foreColor: c1 backColor: c2
	"Answer an instance of me with text set to aText and style set to 
	aTextStyle, composition rectangle is compRect and the clipping rectangle 
	is clipRect."
	| para |
	para := super new.
	para setWithText: aText
		style: aTextStyle
		compositionRectangle: compRect
		clippingRectangle: clipRect
		foreColor: c1 backColor: c2.
	^para! !

!Paragraph methodsFor: 'converting' stamp: 'yo 6/23/2003 19:05'!
asForm
	"Answer a Form made up of the bits that represent the receiver's displayable text."
	| theForm oldBackColor oldForeColor |
	textStyle isTTCStyle ifTrue: [
		theForm :=  (Form extent: compositionRectangle extent depth: 32)
		offset: offset.
	] ifFalse: [
		theForm := (ColorForm extent: compositionRectangle extent)
			offset: offset;
			colors: (Array
				with: (backColor == nil ifTrue: [Color transparent] ifFalse: [backColor])
				with: (foreColor == nil ifTrue: [Color black] ifFalse: [foreColor])).
	].
	oldBackColor := backColor.
	oldForeColor := foreColor.
	backColor := Color white.
	foreColor := Color black.
	self displayOn: theForm
		at: 0@0
		clippingBox: theForm boundingBox
		rule: Form over
		fillColor: nil.
	backColor := oldBackColor.
	foreColor := oldForeColor.
	^ theForm

"Example:
| p |
p := 'Abc' asParagraph.
p foregroundColor: Color red backgroundColor: Color black.
p asForm displayOn: Display at: 30@30 rule: Form over"
! !

!Paragraph methodsFor: 'converting' stamp: ''!
asString
	"Answer the string of characters of the receiver's text."

	^text string! !

!Paragraph methodsFor: 'converting' stamp: ''!
asText
	"Answer the receiver's text."

	^text! !

!Paragraph methodsFor: 'private' stamp: 'MarcusDenker 2/21/2010 08:50'!
awaitMouseUpIn: box repeating: doBlock ifSucceed: succBlock
	"The mouse has gone down in box; track the mouse, inverting the box while it's within, and if, on mouse up, the cursor was still within the box, execute succBlock.  While waiting for the mouse to come up, repeatedly execute doBlock. 5/11/96 sw
	6/10/96 sw: call new method that adds extra feature"

	| p inside lightForm darkForm isLight |

	p := Sensor cursorPoint.
	inside := box insetBy: 1.
	isLight := true.
	lightForm := Form fromDisplay: inside.
	darkForm := lightForm deepCopy reverse.
	[Sensor anyButtonPressed] whileTrue:
		[doBlock value.
		(box containsPoint: (p := Sensor cursorPoint))
			ifTrue: [
					isLight ifTrue: 
						[isLight := false.
						darkForm displayAt: inside origin]]
			ifFalse: [isLight ifFalse:
						[isLight := true.
						lightForm displayAt: inside origin]]].
	(box containsPoint: p)
		ifTrue: [lightForm displayAt: inside origin.
				^ succBlock value]
! !

!Paragraph methodsFor: 'accessing' stamp: ''!
backgroundColor
	backColor == nil ifTrue: [^ Color white].
	^ backColor! !

!Paragraph methodsFor: 'private' stamp: ''!
bottomAtLineIndex: lineIndex 
	"Answer the bottom y of given line."
	| y |
	y := compositionRectangle top.
	lastLine = 0 ifTrue: [^ y + textStyle lineGrid].
	1 to: (lineIndex min: lastLine) do:
		[:i | y := y + (lines at: i) lineHeight].
	^ y
! !

!Paragraph methodsFor: 'display box access' stamp: ''!
boundingBox

	^offset extent: compositionRectangle extent! !

!Paragraph methodsFor: 'selecting' stamp: 'ar 5/28/2000 12:10'!
caretFormForDepth: depth
	"Return a caret form for the given depth."
	"(Paragraph new caretFormForDepth: Display depth) displayOn: Display at: 0@0 rule: Form reverse"

	| box f bb map |
	box := CaretForm boundingBox.
	f := Form extent: box extent depth: depth.
	map := (Color cachedColormapFrom: CaretForm depth to: depth) copy.
	map at: 1 put: (Color transparent pixelValueForDepth: depth).
	map at: 2 put: (Color quickHighLight: depth) first.  "pixel value for reversing"
	bb := BitBlt current toForm: f.
	bb
		sourceForm: CaretForm;
		sourceRect: box;
		destOrigin: 0@0;
		colorMap: map;
 		combinationRule: Form over;
		copyBits.
	^ f! !

!Paragraph methodsFor: 'alignment' stamp: ''!
centered 
	"Set the alignment for the style with which the receiver displays its text 
	so that text is centered in the composition rectangle."

	textStyle alignment: Centered! !

!Paragraph methodsFor: 'character location' stamp: 'ar 5/18/2000 18:33'!
characterBlockAtPoint: aPoint 
	"Answer a CharacterBlock for characters in the text at point aPoint. It is 
	assumed that aPoint has been transformed into coordinates appropriate to 
	the receiver's destinationForm rectangle and the compositionRectangle."

	^CharacterBlockScanner new characterBlockAtPoint: aPoint in: self! !

!Paragraph methodsFor: 'character location' stamp: 'ar 5/18/2000 18:33'!
characterBlockForIndex: targetIndex 
	"Answer a CharacterBlock for character in the text at targetIndex. The 
	coordinates in the CharacterBlock will be appropriate to the intersection 
	of the destinationForm rectangle and the compositionRectangle."

	^CharacterBlockScanner new characterBlockForIndex: targetIndex in: self! !

!Paragraph methodsFor: 'utilities' stamp: ''!
clearVisibleRectangle 
	"Display the area in which the receiver presents its text so that the area 
	is all one tone--in this case, all white."

	destinationForm
	  fill: clippingRectangle
	  rule: rule
	  fillColor: self backgroundColor! !

!Paragraph methodsFor: 'selecting' stamp: 'MarcusDenker 2/21/2010 08:50'!
clickAt: clickPoint for: model controller: aController
	"Give sensitive text a chance to fire.  Display flash: (100@100 extent: 100@100)."
	| startBlock action |
	action := false.
	startBlock := self characterBlockAtPoint: clickPoint.
	(text attributesAt: startBlock stringIndex forStyle: textStyle) 
		do: [:att | | boxes range box | att mayActOnClick ifTrue:
				[range := text rangeOf: att startingAt: startBlock stringIndex.
				boxes := self selectionRectsFrom: (self characterBlockForIndex: range first) 
							to: (self characterBlockForIndex: range last+1).
				box := boxes detect: [:each | each containsPoint: clickPoint]
							ifNone: [^ action].
				self awaitMouseUpIn: box repeating: []
					ifSucceed: [aController terminateAndInitializeAround:
								[(att actOnClickFor: model in: self at: clickPoint editor: aController) ifTrue: [action := true]]]]].
	^ action! !

!Paragraph methodsFor: 'accessing' stamp: ''!
clippingRectangle 
	"Answer the rectangle, defined in absolute coordinates, whose 
	intersection with the destinationForm is the area in which the characters 
	are constrained to display."

	^clippingRectangle! !

!Paragraph methodsFor: 'accessing' stamp: 'di 10/5/97 15:33'!
clippingRectangle: clipRect 
	clippingRectangle := clipRect! !

!Paragraph methodsFor: 'composition' stamp: 'yo 1/23/2003 22:47'!
composeAll
	"Compose a collection of characters into a collection of lines."

	| startIndex stopIndex lineIndex maximumRightX compositionScanner |
	lines := Array new: 32.
	lastLine := 0.
	maximumRightX := 0.
	text size = 0
		ifTrue:
			[compositionRectangle := compositionRectangle withHeight: 0.
			^maximumRightX].
	startIndex := lineIndex := 1.
	stopIndex := text size.
	compositionScanner := MultiCompositionScanner new forParagraph: self.
	[startIndex > stopIndex] whileFalse: 
		[self lineAt: lineIndex 
				put: (compositionScanner composeLine: lineIndex 
										fromCharacterIndex: startIndex 
										inParagraph: self).
		 maximumRightX := compositionScanner rightX max: maximumRightX.
		 startIndex := (lines at: lineIndex) last + 1.
		 lineIndex := lineIndex + 1].
	self updateCompositionHeight.
	self trimLinesTo: lineIndex - 1.
	^ maximumRightX! !

!Paragraph methodsFor: 'accessing' stamp: ''!
compositionRectangle
	"Answer the rectangle whose width is the dimension, modified by 
	indents and tabsLevels, against which line wraparound is measured. The 
	height of the compositionRectangle is reset each time recomposition is 
	required."

	^compositionRectangle! !

!Paragraph methodsFor: 'accessing' stamp: ''!
compositionRectangle: compRectangle 
	"Set the rectangle whose width is the dimension, modified by indents and 
	tabsLevels, against which line wraparound is measured."

	compositionRectangle := compRectangle.
	self composeAll! !

!Paragraph methodsFor: 'private' stamp: 'tk 9/30/96'!
compositionRectangle: compositionRect text: aText style: aTextStyle offset: aPoint

	compositionRectangle := compositionRect copy.
	text := aText.
	textStyle := aTextStyle.
	rule := DefaultRule.
	mask := nil.		"was DefaultMask "
	marginTabsLevel := 0.
	destinationForm := Display.
	offset := aPoint.
	^self composeAll! !

!Paragraph methodsFor: 'private' stamp: ''!
compositionRectangleDelta
	"A handy number -- mostly for scrolling."

	^compositionRectangle top - clippingRectangle top! !

!Paragraph methodsFor: 'display box access' stamp: ''!
computeBoundingBox

	^offset extent: compositionRectangle extent! !

!Paragraph methodsFor: 'utilities' stamp: ''!
deepCopy
	"Don't want to copy the destForm (Display) or fonts in the TextStyle.  9/13/96 tk"

	| new |
	new := self copy.
	new textStyle: textStyle copy.
	new destinationForm: destinationForm.
	new lines: lines copy.
	new text: text deepCopy.
	^ new! !

!Paragraph methodsFor: 'character location' stamp: 'di 10/5/1998 12:59'!
defaultCharacterBlock
	^ CharacterBlock new stringIndex: 1 text: text
			topLeft: compositionRectangle topLeft extent: 0 @ 0! !

!Paragraph methodsFor: 'accessing' stamp: ''!
destinationForm 
	 "Answer the Form into which the characters are scanned."

	^destinationForm! !

!Paragraph methodsFor: 'utilities' stamp: ''!
destinationForm: destForm
	destinationForm := destForm! !

!Paragraph methodsFor: 'private' stamp: ''!
displayLines: linesInterval 
	^ self displayLines: linesInterval
		affectedRectangle: self visibleRectangle! !

!Paragraph methodsFor: 'private' stamp: 'yo 1/23/2003 22:48'!
displayLines: linesInterval affectedRectangle: affectedRectangle
	"This is the first level workhorse in the display portion of the TextForm routines.
	It checks to see which lines in the interval are actually visible, has the
	CharacterScanner display only those, clears out the areas in which display will
	occur, and clears any space remaining in the visibleRectangle following the space
	occupied by lastLine."

	| lineGrid topY firstLineIndex lastLineIndex lastLineIndexBottom |

	"Save some time by only displaying visible lines"
	firstLineIndex := self lineIndexOfTop: affectedRectangle top.
	firstLineIndex < linesInterval first ifTrue: [firstLineIndex := linesInterval first].
	lastLineIndex := self lineIndexOfTop: affectedRectangle bottom - 1.
	lastLineIndex > linesInterval last ifTrue:
			[linesInterval last > lastLine
		 		ifTrue: [lastLineIndex := lastLine]
		  		ifFalse: [lastLineIndex := linesInterval last]].
	lastLineIndexBottom := (self bottomAtLineIndex: lastLineIndex).
	((Rectangle 
		origin: affectedRectangle left @ (topY := self topAtLineIndex: firstLineIndex) 
		corner: affectedRectangle right @ lastLineIndexBottom)
	  intersects: affectedRectangle)
		ifTrue: [ " . . . (skip to clear-below if no lines displayed)"
				MultiDisplayScanner new
					displayLines: (firstLineIndex to: lastLineIndex)
					in: self clippedBy: affectedRectangle].
	lastLineIndex = lastLine ifTrue: 
		 [destinationForm  "Clear out white space below last line"
		 	fill: (affectedRectangle left @ (lastLineIndexBottom max: affectedRectangle top)
				corner: affectedRectangle bottomRight)
		 	rule: rule fillColor: self backgroundColor]! !

!Paragraph methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium
	"Because Paragraphs cache so much information, computation is avoided
	and displayAt: 0@0 is not appropriate here."

	self displayOn: aDisplayMedium
		at: compositionRectangle topLeft
		clippingBox: clippingRectangle
		rule: rule
		fillColor: mask! !

!Paragraph methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aPoint
	"Use internal clippingRect; destination cliping is done during actual display."

	self displayOn: aDisplayMedium at: aPoint
		clippingBox: (clippingRectangle translateBy: aPoint - compositionRectangle topLeft)
		rule: rule fillColor: mask! !

!Paragraph methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm
	"Default display message when aDisplayPoint is in absolute screen
	coordinates."

	rule := ruleInteger.
	mask := aForm.
	clippingRectangle := clipRectangle.
	compositionRectangle := aDisplayPoint extent: compositionRectangle extent.
	(lastLine == nil or: [lastLine < 1]) ifTrue: [self composeAll].
	self displayOn: aDisplayMedium lines: (1 to: lastLine)! !

!Paragraph methodsFor: 'private' stamp: ''!
displayOn: aDisplayMedium lines: lineInterval

	| saveDestinationForm |
	saveDestinationForm := destinationForm.
	destinationForm := aDisplayMedium.
	self displayLines: lineInterval.
	destinationForm := saveDestinationForm! !

!Paragraph methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle align: alignmentPoint with: relativePoint rule: ruleInteger fillColor: aForm 

	self				"Assumes offset has been set!!!!!!!!!!"
	  displayOn: aDisplayMedium
	  at: (offset 
			+ (displayTransformation applyTo: relativePoint) 
			- alignmentPoint) rounded
	  clippingBox: clipRectangle
	  rule: ruleInteger
	  fillColor: aForm.
	! !

!Paragraph methodsFor: 'selecting' stamp: ''!
extendSelectionAt: beginBlock endBlock: endBlock 
	"Answer with an Array of two CharacterBlocks that represent the text 
	selection that the user makes."
	
	(self characterBlockAtPoint: Sensor cursorPoint) <= beginBlock
		ifTrue: [^self mouseMovedFrom: beginBlock 
					pivotBlock: endBlock
					showingCaret: (beginBlock = endBlock)]
		ifFalse: [^self mouseMovedFrom: endBlock 
					pivotBlock: beginBlock
					showingCaret: (beginBlock = endBlock)]
! !

!Paragraph methodsFor: 'selecting' stamp: 'th 9/19/2002 17:27'!
extendSelectionMark: markBlock pointBlock: pointBlock 
	"Answer with an Array of two CharacterBlocks that represent the text 
	selection that the user makes."
	true 
		ifTrue:[^self mouseMovedFrom: pointBlock
					pivotBlock: markBlock
					showingCaret:(pointBlock = markBlock)]
		ifFalse:
		[	| beginBlock endBlock |
			beginBlock := markBlock min: pointBlock.
			endBlock := markBlock max: endBlock.
	
			(self characterBlockAtPoint: Sensor cursorPoint) <= beginBlock
				ifTrue: [^self mouseMovedFrom: beginBlock 
							pivotBlock: endBlock
							showingCaret: (beginBlock = endBlock)]
				ifFalse: [^self mouseMovedFrom: endBlock 
							pivotBlock: beginBlock
							showingCaret: (beginBlock = endBlock)]
		]
! !

!Paragraph methodsFor: 'accessing' stamp: ''!
fillColor 
	"Answer the Form with which each character is combined by the scanner 
	before applying the rule for display."

	^mask! !

!Paragraph methodsFor: 'accessing' stamp: ''!
fillColor: maskForm 
	"Set the argument, maskForm, to be the form with which each character 
	is combined by the scanner before applying the rule for display."

	mask := maskForm! !

!Paragraph methodsFor: 'utilities' stamp: ''!
fit
	"Make the bounding rectangle of the receiver contain all the text without 
	changing the width of the receiver's composition rectangle."

	[(self lineIndexOfTop: clippingRectangle top) = 1]
		whileFalse: [self scrollBy: (0-1)*textStyle lineGrid].
	self updateCompositionHeight.
	clippingRectangle := clippingRectangle withBottom: compositionRectangle bottom! !

!Paragraph methodsFor: 'indicating' stamp: ''!
flash 
	"Complement twice the visible area in which the receiver displays."

	Display flash: clippingRectangle! !

!Paragraph methodsFor: 'accessing' stamp: ''!
height 
	"Answer the height of the composition rectangle."

	^compositionRectangle height! !

!Paragraph methodsFor: 'selecting' stamp: 'jm 7/1/1999 12:31'!
hiliteRect: rect

	| highlightColor |
	highlightColor := Color quickHighLight: destinationForm depth.
	rect ifNotNil: [
		destinationForm
			fill: rect
			rule: Form reverse
			fillColor: highlightColor.
		"destinationForm
			fill: (rect translateBy: 1@1)
			rule: Form reverse
			fillColor: highlightColor" ].
! !

!Paragraph methodsFor: 'accessing' stamp: ''!
indentationOfLineIndex: lineIndex ifBlank: aBlock
	"Answer the number of leading tabs in the line at lineIndex.  If there are
	 no visible characters, pass the number of tabs to aBlock and return its value.
	 If the line is word-wrap overflow, back up a line and recur."

	| arrayIndex first last reader leadingTabs lastSeparator cr tab ch |
	cr := Character cr.
	tab := Character tab.
	arrayIndex := lineIndex.
	[first := (lines at: arrayIndex) first.
	 first > 1 and: [(text string at: first - 1) ~~ cr]] whileTrue: "word wrap"
		[arrayIndex := arrayIndex - 1].
	last := (lines at: lastLine) last.
	reader := ReadStream on: text string from: first to: last.
	leadingTabs := 0.
	[reader atEnd not and: [(ch := reader next) == tab]]
		whileTrue: [leadingTabs := leadingTabs + 1].
	lastSeparator := first - 1 + leadingTabs.
	[reader atEnd not and: [ch isSeparator and: [ch ~~ cr]]]
		whileTrue: [lastSeparator := lastSeparator + 1. ch := reader next].
	lastSeparator = last | (ch == cr)
		ifTrue: [^aBlock value: leadingTabs].
	^leadingTabs! !

!Paragraph methodsFor: 'alignment' stamp: ''!
justified 
	"Set the alignment for the style with which the receiver displays its text 
	so that the characters in each of text end on an even border in the 
	composition rectangle."

	textStyle alignment: Justified! !

!Paragraph methodsFor: 'alignment' stamp: ''!
leftFlush 
	"Set the alignment for the style with which the receiver displays its text 
	so that the characters in each of text begin on an even border in the 
	composition rectangle. This is also known as ragged-right."

	textStyle alignment: LeftFlush! !

!Paragraph methodsFor: 'private' stamp: ''!
leftMarginForCompositionForLine: lineIndex 
	"Build the left margin for composition of a line. Depends upon
	marginTabsLevel and the indent."

	| indent |
	lineIndex = 1
		ifTrue: [indent := textStyle firstIndent]
		ifFalse: [indent := textStyle restIndent].
	^indent + (textStyle leftMarginTabAt: marginTabsLevel)! !

!Paragraph methodsFor: 'private' stamp: 'ar 12/15/2001 23:29'!
leftMarginForDisplayForLine: lineIndex alignment: alignment
	"Build the left margin for display of a line. Depends upon
	leftMarginForComposition, compositionRectangle left and the alignment."

	| pad |
	(alignment = LeftFlush or: [alignment = Justified])
		ifTrue: 
			[^compositionRectangle left 
				+ (self leftMarginForCompositionForLine: lineIndex)].
	"When called from character location code and entire string has been cut,
	there are no valid lines, hence following nil check."
	(lineIndex <= lines size and: [(lines at: lineIndex) notNil])
		ifTrue: 
			[pad := (lines at: lineIndex) paddingWidth]
		ifFalse: 
			[pad := 
				compositionRectangle width - textStyle firstIndent - textStyle rightIndent].
	alignment = Centered 
		ifTrue: 
			[^compositionRectangle left 
				+ (self leftMarginForCompositionForLine: lineIndex) + (pad // 2)].
	alignment = RightFlush 
		ifTrue:
			[^compositionRectangle left 
				+ (self leftMarginForCompositionForLine: lineIndex) + pad].
	self error: ['no such alignment']! !

!Paragraph methodsFor: 'private' stamp: ''!
lineAt: indexInteger put: aTextLineInterval 
	"Store a line, track last, and grow lines if necessary."
	indexInteger > lastLine ifTrue: [lastLine := indexInteger].
	lastLine > lines size ifTrue: [lines := lines , (Array new: lines size)].
	^lines at: indexInteger put: aTextLineInterval! !

!Paragraph methodsFor: 'private' stamp: ''!
lineIndexOfCharacterIndex: characterIndex 
	"Answer the line index for a given characterIndex."

	1 to: lastLine do: 
		[:lineIndex | 
		(lines at: lineIndex) last >= characterIndex ifTrue: [^lineIndex]].
	^lastLine! !

!Paragraph methodsFor: 'private' stamp: ''!
lineIndexOfTop: top 
	"Answer the line index at a given top y."
	| y line |
	lastLine = 0 ifTrue: [^ 1].
	y := compositionRectangle top.
	1 to: lastLine do:
		[:i | line := lines at: i.
		(y := y + line lineHeight) > top ifTrue: [^ i]].
	^ lastLine
! !

!Paragraph methodsFor: 'private' stamp: ''!
lines

	^lines! !

!Paragraph methodsFor: 'utilities' stamp: ''!
lines: lineArray
	lines := lineArray! !

!Paragraph methodsFor: 'accessing' stamp: ''!
mask 
	"Answer the Form with which each character is combined by the scanner 
	before applying the rule for display."

	^mask! !

!Paragraph methodsFor: 'selecting' stamp: 'jm 7/8/97 12:25'!
mouseMovedFrom: beginBlock pivotBlock: pivotBlock showingCaret: caretOn 
	| startBlock stopBlock showingCaret |
	stopBlock := startBlock := beginBlock.
	showingCaret := caretOn.
	[Sensor redButtonPressed]
		whileTrue: 
			[stopBlock := self characterBlockAtPoint: Sensor cursorPoint.
			stopBlock = startBlock
				ifFalse: 
					[showingCaret
						ifTrue: 
							[showingCaret := false.
							self reverseFrom: pivotBlock to: pivotBlock].
			((startBlock >= pivotBlock and: [stopBlock >= pivotBlock])
				or: [startBlock <= pivotBlock and: [stopBlock <= pivotBlock]])
				ifTrue: 
					[self reverseFrom: startBlock to: stopBlock.
					startBlock := stopBlock]
				ifFalse: 
					[self reverseFrom: startBlock to: pivotBlock.
					self reverseFrom: pivotBlock to: stopBlock.
					startBlock := stopBlock].
			(clippingRectangle containsRect: stopBlock) ifFalse:
				[stopBlock top < clippingRectangle top
				ifTrue: [self scrollBy: stopBlock top - clippingRectangle top
						withSelectionFrom: pivotBlock to: stopBlock]
				ifFalse: [self scrollBy: stopBlock bottom + textStyle lineGrid - clippingRectangle bottom
						withSelectionFrom: pivotBlock to: stopBlock]]]].
	pivotBlock = stopBlock ifTrue:
		[showingCaret ifFalse:  "restore caret"
			[self reverseFrom: pivotBlock to: pivotBlock]].
	^ Array with: pivotBlock with: stopBlock! !

!Paragraph methodsFor: 'selecting' stamp: ''!
mouseSelect
	"Answer with an Array of two CharacterBlocks that represent the text 
	selection that the user makes.  Return quickly if the button is noticed up
	to make double-click more responsive."

	| pivotBlock startBlock stopBlock origPoint stillDown |
	stillDown := Sensor redButtonPressed.
	pivotBlock := startBlock := stopBlock :=
		self characterBlockAtPoint: (origPoint := Sensor cursorPoint).
	stillDown := stillDown and: [Sensor redButtonPressed].
	self reverseFrom: startBlock to: startBlock.
	[stillDown and: [Sensor cursorPoint = origPoint]] whileTrue:
		[stillDown := Sensor redButtonPressed].
	(stillDown and: [clippingRectangle containsPoint: Sensor cursorPoint])
		ifFalse: [^Array with: pivotBlock with: stopBlock].
	^ self mouseMovedFrom: startBlock 
		pivotBlock: pivotBlock
		showingCaret: true! !

!Paragraph methodsFor: 'selecting' stamp: ''!
mouseSelect: clickPoint 
	"Track text selection and answer with an Array of two CharacterBlocks."
	| startBlock |
	startBlock := self characterBlockAtPoint: clickPoint.
	self reverseFrom: startBlock to: startBlock.
	^ self mouseMovedFrom: startBlock 
		pivotBlock: startBlock
		showingCaret: true! !

!Paragraph methodsFor: 'private' stamp: ''!
moveBy: delta
	compositionRectangle := compositionRectangle translateBy: delta.
	clippingRectangle := clippingRectangle translateBy: delta.
! !

!Paragraph methodsFor: 'accessing' stamp: ''!
numberOfLines 
	"Answer the number of lines of text in the receiver."

	^lastLine! !

!Paragraph methodsFor: 'indicating' stamp: ''!
outline 
	"Display a border around the visible area in which the receiver presents 
	its text."

	clippingRectangle bottom <= compositionRectangle bottom
	  ifTrue: [Display 
				border: (clippingRectangle intersect: compositionRectangle) 
				width: 2]
	  ifFalse: [Display 
				border: (clippingRectangle intersect: destinationForm boundingBox)
				width: 2].
	! !

!Paragraph methodsFor: 'accessing' stamp: 'ar 5/18/2000 18:34'!
replaceFrom: start to: stop with: aText displaying: displayBoolean
	"Replace the receiver's text starting at position start, stopping at stop, by 
	the characters in aText. It is expected that most requirements for 
	modifications to the receiver will call this code. Certainly all cut's or 
	paste's." 

	| compositionScanner obsoleteLines obsoleteLastLine firstLineIndex lastLineIndex
	startLine stopLine replacementRange visibleRectangle startIndex newLine done
	newStop obsoleteY newY moveRectangle |

	text replaceFrom: start to: stop with: aText.		"Update the text."
	lastLine = 0 ifTrue:
		["if lines have never been set up, measure them and display
		all the lines falling in the visibleRectangle"
		self composeAll.
		displayBoolean ifTrue: [^ self displayLines: (1 to: lastLine)]].

	"save -- things get pretty mashed as we go along"
	obsoleteLines := lines copy.
	obsoleteLastLine := lastLine.

	"find the starting and stopping lines"
	firstLineIndex := startLine := self lineIndexOfCharacterIndex: start.
	stopLine := self lineIndexOfCharacterIndex: stop.

	"how many characters being inserted or deleted
		-- negative if aText size is < characterInterval size."
	replacementRange := aText size - (stop - start + 1).
	"Give ourselves plenty of elbow room."
	compositionRectangle := compositionRectangle withHeight: (textStyle lineGrid * 9999).
	"build a boundingBox of the actual screen space in question -- we'll need it later"
	visibleRectangle := (clippingRectangle intersect: compositionRectangle)
							intersect: destinationForm boundingBox.
	compositionScanner := CompositionScanner new forParagraph: self.		"Initialize a scanner."

	"If the starting line is not also the first line, then measuring must commence from line preceding the one in which characterInterval start appears.  For example, deleting a line with only a carriage return may move characters following the deleted portion of text into the line preceding the deleted line."
	startIndex := (lines at: firstLineIndex) first.
	startLine > 1
		ifTrue: 	[newLine := compositionScanner composeLine: startLine - 1
						fromCharacterIndex: (lines at: startLine - 1) first
						inParagraph: self.
				(lines at: startLine - 1) = newLine
					ifFalse:	["start in line preceding the one with the starting character"
							startLine := startLine - 1.
							self lineAt: startLine put: newLine.
							startIndex := newLine last + 1]].
	startIndex > text size ifTrue:
		["nil lines after a deletion -- remeasure last line below"
		self trimLinesTo: (firstLineIndex - 1 max: 0).
		text size = 0 ifTrue:
			["entire text deleted -- clear visibleRectangle and return."
			displayBoolean ifTrue: [destinationForm fill: visibleRectangle rule: rule fillColor: self backgroundColor].
			self updateCompositionHeight.
			^self]].

	"Now we really get to it."
	done := false.
	lastLineIndex := stopLine.
	[done or: [startIndex > text size]]
		whileFalse: 
		[self lineAt: firstLineIndex put:
			(newLine := compositionScanner composeLine: firstLineIndex
							fromCharacterIndex: startIndex inParagraph: self).
		[(lastLineIndex > obsoleteLastLine
			or: ["no more old lines to compare with?"
				newLine last <
					(newStop := (obsoleteLines at: lastLineIndex) last + replacementRange)])
			  	or: [done]]
			whileFalse: 
			[newStop = newLine last
				ifTrue:	["got the match"
						"get source and dest y's for moving the unchanged lines"
						obsoleteY := self topAtLineIndex: lastLineIndex + 1
									using: obsoleteLines and: obsoleteLastLine.
						newY := self topAtLineIndex: firstLineIndex + 1.
						stopLine := firstLineIndex.
						done := true.
							"Fill in the new line vector with the old unchanged lines.
							Update their starting and stopping indices on the way."
						((lastLineIndex := lastLineIndex + 1) to: obsoleteLastLine) do:
							[:upDatedIndex | 
							self lineAt: (firstLineIndex := firstLineIndex + 1) 
								put: ((obsoleteLines at: upDatedIndex)
							  		slide: replacementRange)].
							"trim off obsolete lines, if any"
						self trimLinesTo: firstLineIndex]
				ifFalse:	[lastLineIndex := lastLineIndex + 1]].
		startIndex := newLine last + 1.
		firstLineIndex := firstLineIndex + 1].

	"Now the lines are up to date -- Whew!!.  What remains is to move
	the 'unchanged' lines and display those which have changed."
	displayBoolean   "Not much to do if not displaying"
		ifFalse: [^ self updateCompositionHeight].
	startIndex > text size ifTrue:
		["If at the end of previous lines simply display lines from the line in
		which the first character of the replacement occured through the
		end of the paragraph."
		self updateCompositionHeight.
		self displayLines:
			(startLine to: (stopLine := firstLineIndex min: lastLine)).
		destinationForm  "Clear out area at the bottom"
			fill: ((visibleRectangle left @ (self topAtLineIndex: lastLine + 1)
						extent: visibleRectangle extent)
					intersect: visibleRectangle)
			rule: rule fillColor: self backgroundColor]
		ifFalse:
		[newY ~= obsoleteY ifTrue:
			["Otherwise first move the unchanged lines within
			the visibleRectangle with a good old bitblt."
			moveRectangle :=
				visibleRectangle left @ (obsoleteY max: visibleRectangle top)
					corner: visibleRectangle corner.
			destinationForm copyBits: moveRectangle from: destinationForm
				at: moveRectangle origin + (0 @ (newY-obsoleteY))
				clippingBox: visibleRectangle
				rule: Form over fillColor: nil].

		"Then display the altered lines."
		self displayLines: (startLine to: stopLine).

		newY < obsoleteY
			ifTrue:
			[(self topAtLineIndex: obsoleteLastLine+1 using: obsoleteLines and: obsoleteLastLine) > visibleRectangle bottom
				ifTrue:
				["A deletion may have 'pulled' previously undisplayed lines
				into the visibleRectangle.  If so, display them."
				self displayLines:
					((self lineIndexOfTop: visibleRectangle bottom - (obsoleteY - newY))
						to: (self lineIndexOfTop: visibleRectangle bottom))].
			"Clear out obsolete material at the bottom of the visibleRectangle."
			destinationForm
				fill: ((visibleRectangle left @ ((self bottomAtLineIndex: lastLine) + 1)
						extent: visibleRectangle extent)
					intersect: visibleRectangle)  "How about just corner: ??"
				rule: rule fillColor: self backgroundColor].

		(newY > obsoleteY and: [obsoleteY < visibleRectangle top])
			ifTrue:
				["An insertion may have 'pushed' previously undisplayed lines
				into the visibleRectangle.  If so, display them."
				self displayLines:
					((self lineIndexOfTop: visibleRectangle top)
						to: (self lineIndexOfTop: visibleRectangle top + (newY-obsoleteY)))].

		self updateCompositionHeight]! !

!Paragraph methodsFor: 'selecting' stamp: ''!
reverseFrom: characterBlock1 to: characterBlock2 
	"Reverse area between the two character blocks given as arguments."
	| visibleRectangle initialRectangle interiorRectangle finalRectangle lineNo baseline caret |
	characterBlock1 = characterBlock2 ifTrue:
		[lineNo := self lineIndexOfCharacterIndex: characterBlock1 stringIndex.
		baseline := lineNo = 0 ifTrue: [textStyle baseline]
							ifFalse: [(lines at: lineNo) baseline].
		caret := self caretFormForDepth: Display depth.
		^ caret  "Use a caret to indicate null selection"
				displayOn: destinationForm
				at: characterBlock1 topLeft + (-3 @ baseline)
				clippingBox: clippingRectangle
				rule: (false "Display depth>8" ifTrue: [9 "not-reverse"]
									ifFalse: [Form reverse])
				fillColor: nil].
	visibleRectangle := 
		(clippingRectangle intersect: compositionRectangle)
			"intersect: destinationForm boundingBox" "not necessary".
	characterBlock1 top = characterBlock2 top
		ifTrue: [characterBlock1 left < characterBlock2 left
					ifTrue: 
						[initialRectangle := 
							(characterBlock1 topLeft corner: characterBlock2 bottomLeft)
								intersect: visibleRectangle]
					ifFalse: 
						[initialRectangle := 
							(characterBlock2 topLeft corner: characterBlock1 bottomLeft)
								intersect: visibleRectangle]]
		ifFalse: [characterBlock1 top < characterBlock2 top
					ifTrue: 
						[initialRectangle := 
							(characterBlock1 topLeft 
								corner: visibleRectangle right @ characterBlock1 bottom)
								intersect: visibleRectangle.
						characterBlock1 bottom = characterBlock2 top
							ifTrue: 
								[finalRectangle := 
									(visibleRectangle left @ characterBlock2 top 
										corner: characterBlock2 bottomLeft)
										intersect: visibleRectangle]
							ifFalse: 
								[interiorRectangle := 
									(visibleRectangle left @ characterBlock1 bottom
										corner: visibleRectangle right 
														@ characterBlock2 top)
										intersect: visibleRectangle.
								finalRectangle := 
									(visibleRectangle left @ characterBlock2 top 
										corner: characterBlock2 bottomLeft)
										intersect: visibleRectangle]]
				ifFalse: 
					[initialRectangle := 
						(visibleRectangle left @ characterBlock1 top 
							corner: characterBlock1 bottomLeft)
							intersect: visibleRectangle.
					characterBlock1 top = characterBlock2 bottom
						ifTrue: 
							[finalRectangle := 
								(characterBlock2 topLeft 
									corner: visibleRectangle right 
												@ characterBlock2 bottom)
									intersect: visibleRectangle]
						ifFalse: 
							[interiorRectangle := 
								(visibleRectangle left @ characterBlock2 bottom 
									corner: visibleRectangle right @ characterBlock1 top)
									intersect: visibleRectangle.
							finalRectangle := 
								(characterBlock2 topLeft 
									corner: visibleRectangle right 
												@ characterBlock2 bottom)
									intersect: visibleRectangle]]].
	self hiliteRect: initialRectangle.
	self hiliteRect: interiorRectangle.
	self hiliteRect: finalRectangle.! !

!Paragraph methodsFor: 'alignment' stamp: ''!
rightFlush 
	"Set the alignment for the style with which the receiver displays its text 
	so that the characters in each of text end on an even border in the 
	composition rectangle but the beginning of each line does not. This is 
	also known as ragged-left."

	textStyle alignment: RightFlush! !

!Paragraph methodsFor: 'private' stamp: ''!
rightMarginForComposition
	"Build the right margin for a line. Depends upon compositionRectangle
	width, marginTabsLevel, and right indent."

	^compositionRectangle width 
		- (textStyle rightMarginTabAt: marginTabsLevel) 
		- textStyle rightIndent! !

!Paragraph methodsFor: 'private' stamp: ''!
rightMarginForDisplay 
	"Build the right margin for a line. Depends upon compositionRectangle
	rightSide, marginTabsLevel, and right indent."

	^compositionRectangle right - 
		textStyle rightIndent - (textStyle rightMarginTabAt: marginTabsLevel)! !

!Paragraph methodsFor: 'accessing' stamp: ''!
rule 
	"Answer the rule according to which character display behaves. For 
	example, rule may equal over, under, reverse."

	^rule! !

!Paragraph methodsFor: 'accessing' stamp: ''!
rule: ruleInteger 
	"Set the rule according to which character display behaves."

	rule := ruleInteger! !

!Paragraph methodsFor: 'scrolling' stamp: ''!
scrollBy: heightToMove 
	^ self scrollBy: heightToMove withSelectionFrom: nil to: nil! !

!Paragraph methodsFor: 'scrolling' stamp: 'hmm 9/16/2000 21:30'!
scrollBy: heightToMove withSelectionFrom: startBlock to: stopBlock 
	"Translate the composition rectangle up (dy<0) by heightToMove.
	Repainting text as necessary, and selection if blocks not nil.
	Return true unless scrolling limits have been reached."
	| max min amount |
	max := 0 max: "cant scroll up more than dist to (top of) bottom line"
		compositionRectangle bottom - textStyle lineGrid - clippingRectangle top.
	min := 0 min: "cant scroll down more than top is above clipRect"
		compositionRectangle top - clippingRectangle top.
	amount := ((heightToMove truncateTo: textStyle lineGrid) min: max) max: min.
	amount ~= 0
		ifTrue: [destinationForm deferUpdatesIn: clippingRectangle while: [
					self scrollUncheckedBy: amount
						withSelectionFrom: startBlock to: stopBlock].
				^ true]
		ifFalse: [^ false]! !

!Paragraph methodsFor: 'scrolling' stamp: ''!
scrollDelta
	"By comparing this before and after, you know if scrolling happened"
	^ clippingRectangle top - compositionRectangle top! !

!Paragraph methodsFor: 'scrolling' stamp: ''!
scrollUncheckedBy: heightToMove withSelectionFrom: startBlock to: stopBlock 
	"Scroll by the given amount.  Copy bits where possible, display the rest.
	If selection blocks are not nil, then select the newly visible text as well."
	| savedClippingRectangle delta |
	delta := 0 @ (0 - heightToMove).
	compositionRectangle := compositionRectangle translateBy: delta.
	startBlock == nil ifFalse:
		[startBlock moveBy: delta.
		stopBlock moveBy: delta].
	savedClippingRectangle := clippingRectangle.
	clippingRectangle := clippingRectangle intersect: Display boundingBox.
	heightToMove abs >= clippingRectangle height
	  ifTrue: 
		["Entire visible region must be repainted"
		self displayLines: (1 to: lastLine) affectedRectangle: clippingRectangle]
	  ifFalse:
		["Copy bits where possible / display the rest"
		destinationForm
			copyBits: clippingRectangle from: destinationForm
			at: clippingRectangle topLeft + delta
			clippingBox: clippingRectangle
			rule: Form over fillColor: nil.
		"Set clippingRectangle to 'vacated' area for lines 'pulled' into view."
		clippingRectangle := heightToMove < 0
			ifTrue:  "On the top"
				[clippingRectangle topLeft corner: clippingRectangle topRight + delta]
			ifFalse:  "At the bottom"
				[clippingRectangle bottomLeft + delta corner: clippingRectangle bottomRight].
		self displayLines: (1 to: lastLine)   "Refresh vacated region"
			affectedRectangle: clippingRectangle].
	startBlock == nil ifFalse:
		[self reverseFrom: startBlock to: stopBlock].
	"And restore the clippingRectangle to its original value. "
	clippingRectangle := savedClippingRectangle! !

!Paragraph methodsFor: 'selecting' stamp: 'di 12/1/97 04:43'!
selectionRectsFrom: characterBlock1 to: characterBlock2 
	"Return an array of rectangles representing the area between the two character blocks given as arguments."
	| visibleRectangle initialRectangle interiorRectangle finalRectangle lineNo baseline |
	characterBlock1 = characterBlock2 ifTrue:
		[lineNo := self lineIndexOfCharacterIndex: characterBlock1 stringIndex.
		baseline := lineNo = 0 ifTrue: [textStyle baseline]
							ifFalse: [(lines at: lineNo) baseline].
		^ Array with: (characterBlock1 topLeft extent: 1 @ baseline)].
	visibleRectangle := clippingRectangle intersect: compositionRectangle.
	characterBlock1 top = characterBlock2 top
		ifTrue: [characterBlock1 left < characterBlock2 left
					ifTrue: 
						[initialRectangle := 
							(characterBlock1 topLeft corner: characterBlock2 bottomLeft)
								intersect: visibleRectangle]
					ifFalse: 
						[initialRectangle := 
							(characterBlock2 topLeft corner: characterBlock1 bottomLeft)
								intersect: visibleRectangle]]
		ifFalse: [characterBlock1 top < characterBlock2 top
					ifTrue: 
						[initialRectangle := 
							(characterBlock1 topLeft 
								corner: visibleRectangle right @ characterBlock1 bottom)
								intersect: visibleRectangle.
						characterBlock1 bottom = characterBlock2 top
							ifTrue: 
								[finalRectangle := 
									(visibleRectangle left @ characterBlock2 top 
										corner: characterBlock2 bottomLeft)
										intersect: visibleRectangle]
							ifFalse: 
								[interiorRectangle := 
									(visibleRectangle left @ characterBlock1 bottom
										corner: visibleRectangle right 
														@ characterBlock2 top)
										intersect: visibleRectangle.
								finalRectangle := 
									(visibleRectangle left @ characterBlock2 top 
										corner: characterBlock2 bottomLeft)
										intersect: visibleRectangle]]
				ifFalse: 
					[initialRectangle := 
						(visibleRectangle left @ characterBlock1 top 
							corner: characterBlock1 bottomLeft)
							intersect: visibleRectangle.
					characterBlock1 top = characterBlock2 bottom
						ifTrue: 
							[finalRectangle := 
								(characterBlock2 topLeft 
									corner: visibleRectangle right 
												@ characterBlock2 bottom)
									intersect: visibleRectangle]
						ifFalse: 
							[interiorRectangle := 
								(visibleRectangle left @ characterBlock2 bottom 
									corner: visibleRectangle right @ characterBlock1 top)
									intersect: visibleRectangle.
							finalRectangle := 
								(characterBlock2 topLeft 
									corner: visibleRectangle right 
												@ characterBlock2 bottom)
									intersect: visibleRectangle]]].
	^ (Array with: initialRectangle with: interiorRectangle with: finalRectangle)
			select: [:rect | rect notNil]! !

!Paragraph methodsFor: 'private' stamp: ''!
setWithText: aText style: aTextStyle 
	"Set text and adjust bounding rectangles to fit."

	| shrink compositionWidth unbounded |
	unbounded := Rectangle origin: 0 @ 0 extent: 9999@9999.
	compositionWidth := self
		setWithText: aText style: aTextStyle compositionRectangle: unbounded clippingRectangle: unbounded.
	compositionRectangle := compositionRectangle withWidth: compositionWidth.
	clippingRectangle := compositionRectangle copy.
	shrink := unbounded width - compositionWidth.
	"Shrink padding widths accordingly"
	1 to: lastLine do:
		[:i | (lines at: i) paddingWidth: (lines at: i) paddingWidth - shrink]! !

!Paragraph methodsFor: 'private' stamp: ''!
setWithText: aText style: aTextStyle compositionRectangle: compRect clippingRectangle: clipRect 
	"Set text and using supplied parameters. Answer max composition width."

	clippingRectangle := clipRect copy.
	^self
		compositionRectangle: compRect
		text: aText
		style: aTextStyle
		offset: 0 @ 0! !

!Paragraph methodsFor: 'private' stamp: ''!
setWithText: aText style: aTextStyle compositionRectangle: compRect clippingRectangle: clipRect foreColor: cf backColor: cb
	"Set text and using supplied parameters. Answer max composition width."

	clippingRectangle := clipRect copy.
	self foregroundColor: cf backgroundColor: cb.
	^ self
		compositionRectangle: compRect
		text: aText
		style: aTextStyle
		offset: 0 @ 0! !

!Paragraph methodsFor: 'accessing' stamp: 'sw 10/29/1999 18:11'!
stringAtLineNumber: aNumber
	(aNumber > lastLine or: [aNumber < 1]) ifTrue: [^ nil].
	^ (text string copyFrom: (lines at: aNumber) first to: (lines at: aNumber) last) copyWithout: Character cr! !

!Paragraph methodsFor: 'accessing' stamp: ''!
text: aText 
	"Set the argument, aText, to be the text for the receiver."

	text := aText.
	self composeAll! !

!Paragraph methodsFor: 'alignment' stamp: ''!
toggleAlignment 
	"Set the alignment for the style with which the receiver displays its text 
	so that it moves from centered to justified to leftFlush to rightFlush and 
	back to centered again."

	textStyle alignment: textStyle alignment + 1! !

!Paragraph methodsFor: 'private' stamp: ''!
topAtLineIndex: lineIndex 
	"Answer the top y of given line."
	| y |
	y := compositionRectangle top.
	lastLine = 0 ifTrue: [lineIndex > 0 ifTrue: [^ y + textStyle lineGrid]. ^ y].
	1 to: (lineIndex-1 min: lastLine) do:
		[:i | y := y + (lines at: i) lineHeight].
	^ y
! !

!Paragraph methodsFor: 'private' stamp: ''!
topAtLineIndex: lineIndex using: otherLines and: otherLastLine
	"Answer the top y of given line."
	| y |
	y := compositionRectangle top.
	otherLastLine = 0 ifTrue: [^ y].
	1 to: (lineIndex-1 min: otherLastLine) do:
		[:i | y := y + (otherLines at: i) lineHeight].
	^ y
! !

!Paragraph methodsFor: 'private' stamp: ''!
trimLinesTo: lastLineInteger

	(lastLineInteger + 1 to: lastLine) do: [:i | lines at: i put: nil].
	(lastLine := lastLineInteger) < (lines size // 2) 
		ifTrue: [lines := lines copyFrom: 1 to: lines size - (lines size // 2)]! !

!Paragraph methodsFor: 'private' stamp: ''!
updateCompositionHeight
	"Mainly used to insure that intersections with compositionRectangle work." 

	compositionRectangle := compositionRectangle withHeight:
		(self bottomAtLineIndex: lastLine) - compositionRectangle top.
	(text size ~= 0 and: [(text at: text size) = CR])
		ifTrue: [compositionRectangle := compositionRectangle withHeight:
					compositionRectangle height + (lines at: lastLine) lineHeight]! !

!Paragraph methodsFor: 'utilities' stamp: ''!
visibleRectangle 
	"May be less than the clippingRectangle if text ends part way down.
	Also some fearful history includes Display intersection;
	it shouldn't be necessary"

	^ (clippingRectangle intersect: compositionRectangle)
		intersect: destinationForm boundingBox! !

!Paragraph methodsFor: 'private' stamp: 'di 8/30/97 11:14'!
withClippingRectangle: clipRect do: aBlock
	| saveClip |
	saveClip := clippingRectangle.
	clippingRectangle := clipRect.
		aBlock value.
	clippingRectangle := saveClip! !

!Paragraph methodsFor: 'composition' stamp: ''!
wrappingBox: compositionRect clippingBox: clippingRect 
	"Set the composition rectangle for the receiver so that the lines wrap 
	within the rectangle, compositionRect, and the display of the text is 
	clipped by the rectangle, clippingRect."

	self compositionRectangle: compositionRect copy
				text: text
				style: textStyle
				offset: offset.
	clippingRectangle := clippingRect copy! !
ParagraphEditor initialize!
